import logging

from .fields import (
    MONTHS,
    WEEK_NUMBERS,
    WEEKDAY_NAMES,
    WEEKDAYS,
    DateOrDateTimeField,
    DateTimeField,
    EnumField,
    EWSElementField,
    IdElementField,
    IntegerField,
    WeekdaysField,
)
from .properties import EWSElement, EWSMeta, IdChangeKeyMixIn, ItemId

log = logging.getLogger(__name__)


def _month_to_str(month):
    return MONTHS[month - 1] if isinstance(month, int) else month


def _weekday_to_str(weekday):
    return WEEKDAYS[weekday - 1] if isinstance(weekday, int) else weekday


def _week_number_to_str(week_number):
    return WEEK_NUMBERS[week_number - 1] if isinstance(week_number, int) else week_number


class Pattern(EWSElement, metaclass=EWSMeta):
    """Base class for all classes implementing recurring pattern elements."""


class Regeneration(Pattern, metaclass=EWSMeta):
    """Base class for all classes implementing recurring regeneration elements."""


class AbsoluteYearlyPattern(Pattern):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/absoluteyearlyrecurrence
    """

    ELEMENT_NAME = "AbsoluteYearlyRecurrence"

    # The day of month of an occurrence, in range 1 -> 31. If a particular month has less days than the day_of_month
    # value, the last day in the month is assumed
    day_of_month = IntegerField(field_uri="DayOfMonth", min=1, max=31, is_required=True)
    # The month of the year, from 1 - 12
    month = EnumField(field_uri="Month", enum=MONTHS, is_required=True)

    def __str__(self):
        return f"Occurs on day {self.day_of_month} of {_month_to_str(self.month)}"


class RelativeYearlyPattern(Pattern):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/relativeyearlyrecurrence
    """

    ELEMENT_NAME = "RelativeYearlyRecurrence"

    # Valid ISO 8601 weekday, as a number in range 1 -> 7 (1 being Monday). The value can also be one of the DAY
    # (or 8), WEEK_DAY (or 9) or WEEKEND_DAY (or 10) consts which is interpreted as the first day, weekday, or weekend
    # day of the year. Despite the field name in EWS, this is not a list.
    weekday = EnumField(field_uri="DaysOfWeek", enum=WEEKDAYS, is_required=True)
    # Week number of the month, in range 1 -> 5. If 5 is specified, this assumes the last week of the month for
    # months that have only 4 weeks
    week_number = EnumField(field_uri="DayOfWeekIndex", enum=WEEK_NUMBERS, is_required=True)
    # The month of the year, from 1 - 12
    month = EnumField(field_uri="Month", enum=MONTHS, is_required=True)

    def __str__(self):
        return (
            f"Occurs on weekday {_weekday_to_str(self.weekday)} in the {_week_number_to_str(self.week_number)} "
            f"week of {_month_to_str(self.month)}"
        )


class AbsoluteMonthlyPattern(Pattern):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/absolutemonthlyrecurrence
    """

    ELEMENT_NAME = "AbsoluteMonthlyRecurrence"

    # Interval, in months, in range 1 -> 99
    interval = IntegerField(field_uri="Interval", min=1, max=99, is_required=True)
    # The day of month of an occurrence, in range 1 -> 31. If a particular month has less days than the day_of_month
    # value, the last day in the month is assumed
    day_of_month = IntegerField(field_uri="DayOfMonth", min=1, max=31, is_required=True)

    def __str__(self):
        return f"Occurs on day {self.day_of_month} of every {self.interval} month(s)"


class RelativeMonthlyPattern(Pattern):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/relativemonthlyrecurrence
    """

    ELEMENT_NAME = "RelativeMonthlyRecurrence"

    # Interval, in months, in range 1 -> 99
    interval = IntegerField(field_uri="Interval", min=1, max=99, is_required=True)
    # Valid ISO 8601 weekday, as a number in range 1 -> 7 (1 being Monday). The value can also be one of the DAY
    # (or 8), WEEK_DAY (or 9) or WEEKEND_DAY (or 10) consts which is interpreted as the first day, weekday, or weekend
    # day of the month. Despite the field name in EWS, this is not a list.
    weekday = EnumField(field_uri="DaysOfWeek", enum=WEEKDAYS, is_required=True)
    # Week number of the month, in range 1 -> 5. If 5 is specified, this assumes the last week of the month for
    # months that have only 4 weeks.
    week_number = EnumField(field_uri="DayOfWeekIndex", enum=WEEK_NUMBERS, is_required=True)

    def __str__(self):
        return (
            f"Occurs on weekday {_weekday_to_str(self.weekday)} in the {_week_number_to_str(self.week_number)} "
            f"week of every {self.interval} month(s)"
        )


class WeeklyPattern(Pattern):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/weeklyrecurrence"""

    ELEMENT_NAME = "WeeklyRecurrence"

    # Interval, in weeks, in range 1 -> 99
    interval = IntegerField(field_uri="Interval", min=1, max=99, is_required=True)
    # List of valid ISO 8601 weekdays, as list of numbers in range 1 -> 7 (1 being Monday)
    weekdays = WeekdaysField(field_uri="DaysOfWeek", enum=WEEKDAY_NAMES, is_required=True)
    # The first day of the week. Defaults to Monday
    first_day_of_week = EnumField(field_uri="FirstDayOfWeek", enum=WEEKDAY_NAMES, default=1, is_required=True)

    def __str__(self):
        weekdays = [_weekday_to_str(i) for i in self.get_field_by_fieldname("weekdays").clean(self.weekdays)]
        return (
            f'Occurs on weekdays {", ".join(weekdays)} of every {self.interval} week(s) where the first day of '
            f"the week is {_weekday_to_str(self.first_day_of_week)}"
        )


class DailyPattern(Pattern):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/dailyrecurrence"""

    ELEMENT_NAME = "DailyRecurrence"

    # Interval, in days, in range 1 -> 999
    interval = IntegerField(field_uri="Interval", min=1, max=999, is_required=True)

    def __str__(self):
        return f"Occurs every {self.interval} day(s)"


class YearlyRegeneration(Regeneration):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/yearlyregeneration"""

    ELEMENT_NAME = "YearlyRegeneration"

    # Interval, in years
    interval = IntegerField(field_uri="Interval", min=1, is_required=True)

    def __str__(self):
        return f"Regenerates every {self.interval} year(s)"


class MonthlyRegeneration(Regeneration):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/monthlyregeneration"""

    ELEMENT_NAME = "MonthlyRegeneration"

    # Interval, in months
    interval = IntegerField(field_uri="Interval", min=1, is_required=True)

    def __str__(self):
        return f"Regenerates every {self.interval} month(s)"


class WeeklyRegeneration(Regeneration):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/weeklyregeneration"""

    ELEMENT_NAME = "WeeklyRegeneration"

    # Interval, in weeks
    interval = IntegerField(field_uri="Interval", min=1, is_required=True)

    def __str__(self):
        return f"Regenerates every {self.interval} week(s)"


class DailyRegeneration(Regeneration):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/dailyregeneration"""

    ELEMENT_NAME = "DailyRegeneration"

    # Interval, in days
    interval = IntegerField(field_uri="Interval", min=1, is_required=True)

    def __str__(self):
        return f"Regenerates every {self.interval} day(s)"


class Boundary(EWSElement, metaclass=EWSMeta):
    """Base class for all classes implementing recurring boundary elements."""


class NoEndPattern(Boundary):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/noendrecurrence"""

    ELEMENT_NAME = "NoEndRecurrence"

    # Start date, as EWSDate or EWSDateTime
    start = DateOrDateTimeField(field_uri="StartDate", is_required=True)

    def __str__(self):
        return f"Starts on {self.start}"


class EndDatePattern(Boundary):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/enddaterecurrence"""

    ELEMENT_NAME = "EndDateRecurrence"

    # Start date, as EWSDate or EWSDateTime
    start = DateOrDateTimeField(field_uri="StartDate", is_required=True)
    # End date, as EWSDate
    end = DateOrDateTimeField(field_uri="EndDate", is_required=True)

    def __str__(self):
        return f"Starts on {self.start}, ends on {self.end}"


class NumberedPattern(Boundary):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/numberedrecurrence"""

    ELEMENT_NAME = "NumberedRecurrence"

    # Start date, as EWSDate or EWSDateTime
    start = DateOrDateTimeField(field_uri="StartDate", is_required=True)
    # The number of occurrences in this pattern, in range 1 -> 999
    number = IntegerField(field_uri="NumberOfOccurrences", min=1, max=999, is_required=True)

    def __str__(self):
        return f"Starts on {self.start} and occurs {self.number} time(s)"


class Occurrence(IdChangeKeyMixIn):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/occurrence"""

    ELEMENT_NAME = "Occurrence"
    ID_ELEMENT_CLS = ItemId

    _id = IdElementField(field_uri="ItemId", value_cls=ID_ELEMENT_CLS)
    # The modified start time of the item, as EWSDateTime
    start = DateTimeField(field_uri="Start")
    # The modified end time of the item, as EWSDateTime
    end = DateTimeField(field_uri="End")
    # The original start time of the item, as EWSDateTime
    original_start = DateTimeField(field_uri="OriginalStart")


# Container elements:
# 'ModifiedOccurrences'
# 'DeletedOccurrences'


class FirstOccurrence(Occurrence):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/firstoccurrence"""

    ELEMENT_NAME = "FirstOccurrence"


class LastOccurrence(Occurrence):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/lastoccurrence"""

    ELEMENT_NAME = "LastOccurrence"


class DeletedOccurrence(EWSElement):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/deletedoccurrence"""

    ELEMENT_NAME = "DeletedOccurrence"

    # The modified start time of the item, as EWSDateTime
    start = DateTimeField(field_uri="Start")


PATTERN_CLASSES = (
    AbsoluteYearlyPattern,
    RelativeYearlyPattern,
    AbsoluteMonthlyPattern,
    RelativeMonthlyPattern,
    WeeklyPattern,
    DailyPattern,
)
REGENERATION_CLASSES = YearlyRegeneration, MonthlyRegeneration, WeeklyRegeneration, DailyRegeneration
BOUNDARY_CLASSES = NoEndPattern, EndDatePattern, NumberedPattern


class Recurrence(EWSElement):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/recurrence-recurrencetype
    """

    ELEMENT_NAME = "Recurrence"
    PATTERN_CLASS_MAP = {cls.response_tag(): cls for cls in PATTERN_CLASSES}
    BOUNDARY_CLASS_MAP = {cls.response_tag(): cls for cls in BOUNDARY_CLASSES}

    pattern = EWSElementField(value_cls=Pattern)
    boundary = EWSElementField(value_cls=Boundary)

    def __init__(self, **kwargs):
        # Allow specifying a start, end and/or number as a shortcut to creating a boundary
        start = kwargs.pop("start", None)
        end = kwargs.pop("end", None)
        number = kwargs.pop("number", None)
        if any([start, end, number]):
            if "boundary" in kwargs:
                raise ValueError("'boundary' is not allowed in combination with 'start', 'end' or 'number'")
            if start and not end and not number:
                kwargs["boundary"] = NoEndPattern(start=start)
            elif start and end and not number:
                kwargs["boundary"] = EndDatePattern(start=start, end=end)
            elif start and number and not end:
                kwargs["boundary"] = NumberedPattern(start=start, number=number)
            else:
                raise ValueError("Unsupported 'start', 'end', 'number' combination")
        super().__init__(**kwargs)

    @classmethod
    def from_xml(cls, elem, account):
        pattern, boundary = None, None
        for child_elem in elem:
            if child_elem.tag in cls.PATTERN_CLASS_MAP:
                pattern = cls.PATTERN_CLASS_MAP[child_elem.tag].from_xml(elem=child_elem, account=account)
            elif child_elem.tag in cls.BOUNDARY_CLASS_MAP:
                boundary = cls.BOUNDARY_CLASS_MAP[child_elem.tag].from_xml(elem=child_elem, account=account)
        return cls(pattern=pattern, boundary=boundary)

    def __str__(self):
        return f"Pattern: {self.pattern}, Boundary: {self.boundary}"


class TaskRecurrence(Recurrence):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/recurrence-taskrecurrencetype
    """

    PATTERN_CLASS_MAP = {cls.response_tag(): cls for cls in PATTERN_CLASSES + REGENERATION_CLASSES}



from ..errors import MalformedResponseError
from ..properties import FailedMailbox, SearchableMailbox
from ..util import MNS, add_xml_child, create_element
from ..version import EXCHANGE_2013
from .common import EWSService


class GetSearchableMailboxes(EWSService):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getsearchablemailboxes-operation
    """

    SERVICE_NAME = "GetSearchableMailboxes"
    element_container_name = f"{{{MNS}}}SearchableMailboxes"
    failed_mailboxes_container_name = f"{{{MNS}}}FailedMailboxes"
    supported_from = EXCHANGE_2013
    cls_map = {cls.response_tag(): cls for cls in (SearchableMailbox, FailedMailbox)}

    def call(self, search_filter, expand_group_membership):
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(
                    search_filter=search_filter,
                    expand_group_membership=expand_group_membership,
                )
            )
        )

    def _elem_to_obj(self, elem):
        return self.cls_map[elem.tag].from_xml(elem=elem, account=None)

    def get_payload(self, search_filter, expand_group_membership):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        if search_filter:
            add_xml_child(payload, "m:SearchFilter", search_filter)
        if expand_group_membership is not None:
            add_xml_child(payload, "m:ExpandGroupMembership", "true" if expand_group_membership else "false")
        return payload

    def _get_elements_in_response(self, response):
        for msg in response:
            for container_name in (self.element_container_name, self.failed_mailboxes_container_name):
                try:
                    container = self._get_element_container(message=msg, name=container_name)
                except MalformedResponseError:
                    # Responses may contain no mailboxes of either kind. _get_element_container() does not accept this.
                    continue
                yield from self._get_elements_in_container(container=container)



from ..folders.base import BaseFolder
from ..util import MNS, create_element
from .common import EWSAccountService, item_ids_element, shape_element


class GetItem(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getitem-operation"""

    SERVICE_NAME = "GetItem"
    element_container_name = f"{{{MNS}}}Items"

    def call(self, items, additional_fields, shape):
        """Return all items in an account that correspond to a list of ID's, in stable order.

        :param items: a list of (id, changekey) tuples or Item objects
        :param additional_fields: the extra fields that should be returned with the item, as FieldPath objects
        :param shape: The shape of returned objects

        :return: XML elements for the items, in stable order
        """
        return self._elems_to_objs(
            self._chunked_get_elements(
                self.get_payload,
                items=items,
                additional_fields=additional_fields,
                shape=shape,
            )
        )

    def _elem_to_obj(self, elem):
        return BaseFolder.item_model_from_tag(elem.tag).from_xml(elem=elem, account=self.account)

    def get_payload(self, items, additional_fields, shape):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        payload.append(
            shape_element(
                tag="m:ItemShape", shape=shape, additional_fields=additional_fields, version=self.account.version
            )
        )
        payload.append(item_ids_element(items=items, version=self.account.version))
        return payload



"""A dict to translate from IANA location name to Windows timezone name. Translations taken from CLDR_WINZONE_URL
"""
import re

import requests

from .util import to_xml

CLDR_WINZONE_URL = "https://raw.githubusercontent.com/unicode-org/cldr/master/common/supplemental/windowsZones.xml"
DEFAULT_TERRITORY = "001"
CLDR_WINZONE_TYPE_VERSION = "2021a"
CLDR_WINZONE_OTHER_VERSION = "7e11800"


def generate_map(timeout=10):
    """Create a new CLDR_TO_MS_TIMEZONE_MAP map from the CLDR data. Used when the CLDR database is updated.

    :param timeout:  (Default value = 10)
    :return:
    """
    r = requests.get(CLDR_WINZONE_URL, timeout=timeout)
    if r.status_code != 200:
        raise ValueError(f"Unexpected response: {r}")
    tz_map = {}
    timezones_elem = to_xml(r.content).find("windowsZones").find("mapTimezones")
    type_version = timezones_elem.get("typeVersion")
    other_version = timezones_elem.get("otherVersion")
    for e in timezones_elem.findall("mapZone"):
        for location in re.split(r"\s+", e.get("type")):
            if e.get("territory") == DEFAULT_TERRITORY or location not in tz_map:
                # Prefer default territory. This is so MS_TIMEZONE_TO_IANA_MAP maps from MS timezone ID back to the
                # "preferred" region/location timezone name.
                tz_map[location] = e.get("other"), e.get("territory")
    return type_version, other_version, tz_map


# This map is generated irregularly from generate_map(). Do not edit manually - make corrections to
# IANA_TO_MS_TIMEZONE_MAP instead. We provide this map to avoid hammering the CLDR_WINZONE_URL.
#
# This list was generated from CLDR_WINZONE_URL version CLDR_WINZONE_VERSION.
CLDR_TO_MS_TIMEZONE_MAP = {
    "Africa/Abidjan": ("Greenwich Standard Time", "CI"),
    "Africa/Accra": ("Greenwich Standard Time", "GH"),
    "Africa/Addis_Ababa": ("E. Africa Standard Time", "ET"),
    "Africa/Algiers": ("W. Central Africa Standard Time", "DZ"),
    "Africa/Asmera": ("E. Africa Standard Time", "ER"),
    "Africa/Bamako": ("Greenwich Standard Time", "ML"),
    "Africa/Bangui": ("W. Central Africa Standard Time", "CF"),
    "Africa/Banjul": ("Greenwich Standard Time", "GM"),
    "Africa/Bissau": ("Greenwich Standard Time", "GW"),
    "Africa/Blantyre": ("South Africa Standard Time", "MW"),
    "Africa/Brazzaville": ("W. Central Africa Standard Time", "CG"),
    "Africa/Bujumbura": ("South Africa Standard Time", "BI"),
    "Africa/Cairo": ("Egypt Standard Time", "001"),
    "Africa/Casablanca": ("Morocco Standard Time", "001"),
    "Africa/Ceuta": ("Romance Standard Time", "ES"),
    "Africa/Conakry": ("Greenwich Standard Time", "GN"),
    "Africa/Dakar": ("Greenwich Standard Time", "SN"),
    "Africa/Dar_es_Salaam": ("E. Africa Standard Time", "TZ"),
    "Africa/Djibouti": ("E. Africa Standard Time", "DJ"),
    "Africa/Douala": ("W. Central Africa Standard Time", "CM"),
    "Africa/El_Aaiun": ("Morocco Standard Time", "EH"),
    "Africa/Freetown": ("Greenwich Standard Time", "SL"),
    "Africa/Gaborone": ("South Africa Standard Time", "BW"),
    "Africa/Harare": ("South Africa Standard Time", "ZW"),
    "Africa/Johannesburg": ("South Africa Standard Time", "001"),
    "Africa/Juba": ("South Sudan Standard Time", "001"),
    "Africa/Kampala": ("E. Africa Standard Time", "UG"),
    "Africa/Khartoum": ("Sudan Standard Time", "001"),
    "Africa/Kigali": ("South Africa Standard Time", "RW"),
    "Africa/Kinshasa": ("W. Central Africa Standard Time", "CD"),
    "Africa/Lagos": ("W. Central Africa Standard Time", "001"),
    "Africa/Libreville": ("W. Central Africa Standard Time", "GA"),
    "Africa/Lome": ("Greenwich Standard Time", "TG"),
    "Africa/Luanda": ("W. Central Africa Standard Time", "AO"),
    "Africa/Lubumbashi": ("South Africa Standard Time", "CD"),
    "Africa/Lusaka": ("South Africa Standard Time", "ZM"),
    "Africa/Malabo": ("W. Central Africa Standard Time", "GQ"),
    "Africa/Maputo": ("South Africa Standard Time", "MZ"),
    "Africa/Maseru": ("South Africa Standard Time", "LS"),
    "Africa/Mbabane": ("South Africa Standard Time", "SZ"),
    "Africa/Mogadishu": ("E. Africa Standard Time", "SO"),
    "Africa/Monrovia": ("Greenwich Standard Time", "LR"),
    "Africa/Nairobi": ("E. Africa Standard Time", "001"),
    "Africa/Ndjamena": ("W. Central Africa Standard Time", "TD"),
    "Africa/Niamey": ("W. Central Africa Standard Time", "NE"),
    "Africa/Nouakchott": ("Greenwich Standard Time", "MR"),
    "Africa/Ouagadougou": ("Greenwich Standard Time", "BF"),
    "Africa/Porto-Novo": ("W. Central Africa Standard Time", "BJ"),
    "Africa/Sao_Tome": ("Sao Tome Standard Time", "001"),
    "Africa/Tripoli": ("Libya Standard Time", "001"),
    "Africa/Tunis": ("W. Central Africa Standard Time", "TN"),
    "Africa/Windhoek": ("Namibia Standard Time", "001"),
    "America/Adak": ("Aleutian Standard Time", "001"),
    "America/Anchorage": ("Alaskan Standard Time", "001"),
    "America/Anguilla": ("SA Western Standard Time", "AI"),
    "America/Antigua": ("SA Western Standard Time", "AG"),
    "America/Araguaina": ("Tocantins Standard Time", "001"),
    "America/Argentina/La_Rioja": ("Argentina Standard Time", "AR"),
    "America/Argentina/Rio_Gallegos": ("Argentina Standard Time", "AR"),
    "America/Argentina/Salta": ("Argentina Standard Time", "AR"),
    "America/Argentina/San_Juan": ("Argentina Standard Time", "AR"),
    "America/Argentina/San_Luis": ("Argentina Standard Time", "AR"),
    "America/Argentina/Tucuman": ("Argentina Standard Time", "AR"),
    "America/Argentina/Ushuaia": ("Argentina Standard Time", "AR"),
    "America/Aruba": ("SA Western Standard Time", "AW"),
    "America/Asuncion": ("Paraguay Standard Time", "001"),
    "America/Bahia": ("Bahia Standard Time", "001"),
    "America/Bahia_Banderas": ("Central Standard Time (Mexico)", "MX"),
    "America/Barbados": ("SA Western Standard Time", "BB"),
    "America/Belem": ("SA Eastern Standard Time", "BR"),
    "America/Belize": ("Central America Standard Time", "BZ"),
    "America/Blanc-Sablon": ("SA Western Standard Time", "CA"),
    "America/Boa_Vista": ("SA Western Standard Time", "BR"),
    "America/Bogota": ("SA Pacific Standard Time", "001"),
    "America/Boise": ("Mountain Standard Time", "US"),
    "America/Buenos_Aires": ("Argentina Standard Time", "001"),
    "America/Cambridge_Bay": ("Mountain Standard Time", "CA"),
    "America/Campo_Grande": ("Central Brazilian Standard Time", "BR"),
    "America/Cancun": ("Eastern Standard Time (Mexico)", "001"),
    "America/Caracas": ("Venezuela Standard Time", "001"),
    "America/Catamarca": ("Argentina Standard Time", "AR"),
    "America/Cayenne": ("SA Eastern Standard Time", "001"),
    "America/Cayman": ("SA Pacific Standard Time", "KY"),
    "America/Chicago": ("Central Standard Time", "001"),
    "America/Chihuahua": ("Mountain Standard Time (Mexico)", "001"),
    "America/Coral_Harbour": ("SA Pacific Standard Time", "CA"),
    "America/Cordoba": ("Argentina Standard Time", "AR"),
    "America/Costa_Rica": ("Central America Standard Time", "CR"),
    "America/Creston": ("US Mountain Standard Time", "CA"),
    "America/Cuiaba": ("Central Brazilian Standard Time", "001"),
    "America/Curacao": ("SA Western Standard Time", "CW"),
    "America/Danmarkshavn": ("Greenwich Standard Time", "GL"),
    "America/Dawson": ("Yukon Standard Time", "CA"),
    "America/Dawson_Creek": ("US Mountain Standard Time", "CA"),
    "America/Denver": ("Mountain Standard Time", "001"),
    "America/Detroit": ("Eastern Standard Time", "US"),
    "America/Dominica": ("SA Western Standard Time", "DM"),
    "America/Edmonton": ("Mountain Standard Time", "CA"),
    "America/Eirunepe": ("SA Pacific Standard Time", "BR"),
    "America/El_Salvador": ("Central America Standard Time", "SV"),
    "America/Fort_Nelson": ("US Mountain Standard Time", "CA"),
    "America/Fortaleza": ("SA Eastern Standard Time", "BR"),
    "America/Glace_Bay": ("Atlantic Standard Time", "CA"),
    "America/Godthab": ("Greenland Standard Time", "001"),
    "America/Goose_Bay": ("Atlantic Standard Time", "CA"),
    "America/Grand_Turk": ("Turks And Caicos Standard Time", "001"),
    "America/Grenada": ("SA Western Standard Time", "GD"),
    "America/Guadeloupe": ("SA Western Standard Time", "GP"),
    "America/Guatemala": ("Central America Standard Time", "001"),
    "America/Guayaquil": ("SA Pacific Standard Time", "EC"),
    "America/Guyana": ("SA Western Standard Time", "GY"),
    "America/Halifax": ("Atlantic Standard Time", "001"),
    "America/Havana": ("Cuba Standard Time", "001"),
    "America/Hermosillo": ("US Mountain Standard Time", "MX"),
    "America/Indiana/Knox": ("Central Standard Time", "US"),
    "America/Indiana/Marengo": ("US Eastern Standard Time", "US"),
    "America/Indiana/Petersburg": ("Eastern Standard Time", "US"),
    "America/Indiana/Tell_City": ("Central Standard Time", "US"),
    "America/Indiana/Vevay": ("US Eastern Standard Time", "US"),
    "America/Indiana/Vincennes": ("Eastern Standard Time", "US"),
    "America/Indiana/Winamac": ("Eastern Standard Time", "US"),
    "America/Indianapolis": ("US Eastern Standard Time", "001"),
    "America/Inuvik": ("Mountain Standard Time", "CA"),
    "America/Iqaluit": ("Eastern Standard Time", "CA"),
    "America/Jamaica": ("SA Pacific Standard Time", "JM"),
    "America/Jujuy": ("Argentina Standard Time", "AR"),
    "America/Juneau": ("Alaskan Standard Time", "US"),
    "America/Kentucky/Monticello": ("Eastern Standard Time", "US"),
    "America/Kralendijk": ("SA Western Standard Time", "BQ"),
    "America/La_Paz": ("SA Western Standard Time", "001"),
    "America/Lima": ("SA Pacific Standard Time", "PE"),
    "America/Los_Angeles": ("Pacific Standard Time", "001"),
    "America/Louisville": ("Eastern Standard Time", "US"),
    "America/Lower_Princes": ("SA Western Standard Time", "SX"),
    "America/Maceio": ("SA Eastern Standard Time", "BR"),
    "America/Managua": ("Central America Standard Time", "NI"),
    "America/Manaus": ("SA Western Standard Time", "BR"),
    "America/Marigot": ("SA Western Standard Time", "MF"),
    "America/Martinique": ("SA Western Standard Time", "MQ"),
    "America/Matamoros": ("Central Standard Time", "MX"),
    "America/Mazatlan": ("Mountain Standard Time (Mexico)", "MX"),
    "America/Mendoza": ("Argentina Standard Time", "AR"),
    "America/Menominee": ("Central Standard Time", "US"),
    "America/Merida": ("Central Standard Time (Mexico)", "MX"),
    "America/Metlakatla": ("Alaskan Standard Time", "US"),
    "America/Mexico_City": ("Central Standard Time (Mexico)", "001"),
    "America/Miquelon": ("Saint Pierre Standard Time", "001"),
    "America/Moncton": ("Atlantic Standard Time", "CA"),
    "America/Monterrey": ("Central Standard Time (Mexico)", "MX"),
    "America/Montevideo": ("Montevideo Standard Time", "001"),
    "America/Montreal": ("Eastern Standard Time", "CA"),
    "America/Montserrat": ("SA Western Standard Time", "MS"),
    "America/Nassau": ("Eastern Standard Time", "BS"),
    "America/New_York": ("Eastern Standard Time", "001"),
    "America/Nipigon": ("Eastern Standard Time", "CA"),
    "America/Nome": ("Alaskan Standard Time", "US"),
    "America/Noronha": ("UTC-02", "BR"),
    "America/North_Dakota/Beulah": ("Central Standard Time", "US"),
    "America/North_Dakota/Center": ("Central Standard Time", "US"),
    "America/North_Dakota/New_Salem": ("Central Standard Time", "US"),
    "America/Ojinaga": ("Mountain Standard Time", "MX"),
    "America/Panama": ("SA Pacific Standard Time", "PA"),
    "America/Pangnirtung": ("Eastern Standard Time", "CA"),
    "America/Paramaribo": ("SA Eastern Standard Time", "SR"),
    "America/Phoenix": ("US Mountain Standard Time", "001"),
    "America/Port-au-Prince": ("Haiti Standard Time", "001"),
    "America/Port_of_Spain": ("SA Western Standard Time", "TT"),
    "America/Porto_Velho": ("SA Western Standard Time", "BR"),
    "America/Puerto_Rico": ("SA Western Standard Time", "PR"),
    "America/Punta_Arenas": ("Magallanes Standard Time", "001"),
    "America/Rainy_River": ("Central Standard Time", "CA"),
    "America/Rankin_Inlet": ("Central Standard Time", "CA"),
    "America/Recife": ("SA Eastern Standard Time", "BR"),
    "America/Regina": ("Canada Central Standard Time", "001"),
    "America/Resolute": ("Central Standard Time", "CA"),
    "America/Rio_Branco": ("SA Pacific Standard Time", "BR"),
    "America/Santa_Isabel": ("Pacific Standard Time (Mexico)", "MX"),
    "America/Santarem": ("SA Eastern Standard Time", "BR"),
    "America/Santiago": ("Pacific SA Standard Time", "001"),
    "America/Santo_Domingo": ("SA Western Standard Time", "DO"),
    "America/Sao_Paulo": ("E. South America Standard Time", "001"),
    "America/Scoresbysund": ("Azores Standard Time", "GL"),
    "America/Sitka": ("Alaskan Standard Time", "US"),
    "America/St_Barthelemy": ("SA Western Standard Time", "BL"),
    "America/St_Johns": ("Newfoundland Standard Time", "001"),
    "America/St_Kitts": ("SA Western Standard Time", "KN"),
    "America/St_Lucia": ("SA Western Standard Time", "LC"),
    "America/St_Thomas": ("SA Western Standard Time", "VI"),
    "America/St_Vincent": ("SA Western Standard Time", "VC"),
    "America/Swift_Current": ("Canada Central Standard Time", "CA"),
    "America/Tegucigalpa": ("Central America Standard Time", "HN"),
    "America/Thule": ("Atlantic Standard Time", "GL"),
    "America/Thunder_Bay": ("Eastern Standard Time", "CA"),
    "America/Tijuana": ("Pacific Standard Time (Mexico)", "001"),
    "America/Toronto": ("Eastern Standard Time", "CA"),
    "America/Tortola": ("SA Western Standard Time", "VG"),
    "America/Vancouver": ("Pacific Standard Time", "CA"),
    "America/Whitehorse": ("Yukon Standard Time", "001"),
    "America/Winnipeg": ("Central Standard Time", "CA"),
    "America/Yakutat": ("Alaskan Standard Time", "US"),
    "America/Yellowknife": ("Mountain Standard Time", "CA"),
    "Antarctica/Casey": ("Central Pacific Standard Time", "AQ"),
    "Antarctica/Davis": ("SE Asia Standard Time", "AQ"),
    "Antarctica/DumontDUrville": ("West Pacific Standard Time", "AQ"),
    "Antarctica/Macquarie": ("Tasmania Standard Time", "AU"),
    "Antarctica/Mawson": ("West Asia Standard Time", "AQ"),
    "Antarctica/McMurdo": ("New Zealand Standard Time", "AQ"),
    "Antarctica/Palmer": ("SA Eastern Standard Time", "AQ"),
    "Antarctica/Rothera": ("SA Eastern Standard Time", "AQ"),
    "Antarctica/Syowa": ("E. Africa Standard Time", "AQ"),
    "Antarctica/Vostok": ("Central Asia Standard Time", "AQ"),
    "Arctic/Longyearbyen": ("W. Europe Standard Time", "SJ"),
    "Asia/Aden": ("Arab Standard Time", "YE"),
    "Asia/Almaty": ("Central Asia Standard Time", "001"),
    "Asia/Amman": ("Jordan Standard Time", "001"),
    "Asia/Anadyr": ("Russia Time Zone 11", "RU"),
    "Asia/Aqtau": ("West Asia Standard Time", "KZ"),
    "Asia/Aqtobe": ("West Asia Standard Time", "KZ"),
    "Asia/Ashgabat": ("West Asia Standard Time", "TM"),
    "Asia/Atyrau": ("West Asia Standard Time", "KZ"),
    "Asia/Baghdad": ("Arabic Standard Time", "001"),
    "Asia/Bahrain": ("Arab Standard Time", "BH"),
    "Asia/Baku": ("Azerbaijan Standard Time", "001"),
    "Asia/Bangkok": ("SE Asia Standard Time", "001"),
    "Asia/Barnaul": ("Altai Standard Time", "001"),
    "Asia/Beirut": ("Middle East Standard Time", "001"),
    "Asia/Bishkek": ("Central Asia Standard Time", "KG"),
    "Asia/Brunei": ("Singapore Standard Time", "BN"),
    "Asia/Calcutta": ("India Standard Time", "001"),
    "Asia/Chita": ("Transbaikal Standard Time", "001"),
    "Asia/Choibalsan": ("Ulaanbaatar Standard Time", "MN"),
    "Asia/Colombo": ("Sri Lanka Standard Time", "001"),
    "Asia/Damascus": ("Syria Standard Time", "001"),
    "Asia/Dhaka": ("Bangladesh Standard Time", "001"),
    "Asia/Dili": ("Tokyo Standard Time", "TL"),
    "Asia/Dubai": ("Arabian Standard Time", "001"),
    "Asia/Dushanbe": ("West Asia Standard Time", "TJ"),
    "Asia/Famagusta": ("GTB Standard Time", "CY"),
    "Asia/Gaza": ("West Bank Standard Time", "PS"),
    "Asia/Hebron": ("West Bank Standard Time", "001"),
    "Asia/Hong_Kong": ("China Standard Time", "HK"),
    "Asia/Hovd": ("W. Mongolia Standard Time", "001"),
    "Asia/Irkutsk": ("North Asia East Standard Time", "001"),
    "Asia/Jakarta": ("SE Asia Standard Time", "ID"),
    "Asia/Jayapura": ("Tokyo Standard Time", "ID"),
    "Asia/Jerusalem": ("Israel Standard Time", "001"),
    "Asia/Kabul": ("Afghanistan Standard Time", "001"),
    "Asia/Kamchatka": ("Russia Time Zone 11", "001"),
    "Asia/Karachi": ("Pakistan Standard Time", "001"),
    "Asia/Katmandu": ("Nepal Standard Time", "001"),
    "Asia/Khandyga": ("Yakutsk Standard Time", "RU"),
    "Asia/Krasnoyarsk": ("North Asia Standard Time", "001"),
    "Asia/Kuala_Lumpur": ("Singapore Standard Time", "MY"),
    "Asia/Kuching": ("Singapore Standard Time", "MY"),
    "Asia/Kuwait": ("Arab Standard Time", "KW"),
    "Asia/Macau": ("China Standard Time", "MO"),
    "Asia/Magadan": ("Magadan Standard Time", "001"),
    "Asia/Makassar": ("Singapore Standard Time", "ID"),
    "Asia/Manila": ("Singapore Standard Time", "PH"),
    "Asia/Muscat": ("Arabian Standard Time", "OM"),
    "Asia/Nicosia": ("GTB Standard Time", "CY"),
    "Asia/Novokuznetsk": ("North Asia Standard Time", "RU"),
    "Asia/Novosibirsk": ("N. Central Asia Standard Time", "001"),
    "Asia/Omsk": ("Omsk Standard Time", "001"),
    "Asia/Oral": ("West Asia Standard Time", "KZ"),
    "Asia/Phnom_Penh": ("SE Asia Standard Time", "KH"),
    "Asia/Pontianak": ("SE Asia Standard Time", "ID"),
    "Asia/Pyongyang": ("North Korea Standard Time", "001"),
    "Asia/Qatar": ("Arab Standard Time", "QA"),
    "Asia/Qostanay": ("Central Asia Standard Time", "KZ"),
    "Asia/Qyzylorda": ("Qyzylorda Standard Time", "001"),
    "Asia/Rangoon": ("Myanmar Standard Time", "001"),
    "Asia/Riyadh": ("Arab Standard Time", "001"),
    "Asia/Saigon": ("SE Asia Standard Time", "VN"),
    "Asia/Sakhalin": ("Sakhalin Standard Time", "001"),
    "Asia/Samarkand": ("West Asia Standard Time", "UZ"),
    "Asia/Seoul": ("Korea Standard Time", "001"),
    "Asia/Shanghai": ("China Standard Time", "001"),
    "Asia/Singapore": ("Singapore Standard Time", "001"),
    "Asia/Srednekolymsk": ("Russia Time Zone 10", "001"),
    "Asia/Taipei": ("Taipei Standard Time", "001"),
    "Asia/Tashkent": ("West Asia Standard Time", "001"),
    "Asia/Tbilisi": ("Georgian Standard Time", "001"),
    "Asia/Tehran": ("Iran Standard Time", "001"),
    "Asia/Thimphu": ("Bangladesh Standard Time", "BT"),
    "Asia/Tokyo": ("Tokyo Standard Time", "001"),
    "Asia/Tomsk": ("Tomsk Standard Time", "001"),
    "Asia/Ulaanbaatar": ("Ulaanbaatar Standard Time", "001"),
    "Asia/Urumqi": ("Central Asia Standard Time", "CN"),
    "Asia/Ust-Nera": ("Vladivostok Standard Time", "RU"),
    "Asia/Vientiane": ("SE Asia Standard Time", "LA"),
    "Asia/Vladivostok": ("Vladivostok Standard Time", "001"),
    "Asia/Yakutsk": ("Yakutsk Standard Time", "001"),
    "Asia/Yekaterinburg": ("Ekaterinburg Standard Time", "001"),
    "Asia/Yerevan": ("Caucasus Standard Time", "001"),
    "Atlantic/Azores": ("Azores Standard Time", "001"),
    "Atlantic/Bermuda": ("Atlantic Standard Time", "BM"),
    "Atlantic/Canary": ("GMT Standard Time", "ES"),
    "Atlantic/Cape_Verde": ("Cape Verde Standard Time", "001"),
    "Atlantic/Faeroe": ("GMT Standard Time", "FO"),
    "Atlantic/Madeira": ("GMT Standard Time", "PT"),
    "Atlantic/Reykjavik": ("Greenwich Standard Time", "001"),
    "Atlantic/South_Georgia": ("UTC-02", "GS"),
    "Atlantic/St_Helena": ("Greenwich Standard Time", "SH"),
    "Atlantic/Stanley": ("SA Eastern Standard Time", "FK"),
    "Australia/Adelaide": ("Cen. Australia Standard Time", "001"),
    "Australia/Brisbane": ("E. Australia Standard Time", "001"),
    "Australia/Broken_Hill": ("Cen. Australia Standard Time", "AU"),
    "Australia/Currie": ("Tasmania Standard Time", "AU"),
    "Australia/Darwin": ("AUS Central Standard Time", "001"),
    "Australia/Eucla": ("Aus Central W. Standard Time", "001"),
    "Australia/Hobart": ("Tasmania Standard Time", "001"),
    "Australia/Lindeman": ("E. Australia Standard Time", "AU"),
    "Australia/Lord_Howe": ("Lord Howe Standard Time", "001"),
    "Australia/Melbourne": ("AUS Eastern Standard Time", "AU"),
    "Australia/Perth": ("W. Australia Standard Time", "001"),
    "Australia/Sydney": ("AUS Eastern Standard Time", "001"),
    "CST6CDT": ("Central Standard Time", "ZZ"),
    "EST5EDT": ("Eastern Standard Time", "ZZ"),
    "Etc/GMT": ("UTC", "ZZ"),
    "Etc/GMT+1": ("Cape Verde Standard Time", "ZZ"),
    "Etc/GMT+10": ("Hawaiian Standard Time", "ZZ"),
    "Etc/GMT+11": ("UTC-11", "001"),
    "Etc/GMT+12": ("Dateline Standard Time", "001"),
    "Etc/GMT+2": ("UTC-02", "001"),
    "Etc/GMT+3": ("SA Eastern Standard Time", "ZZ"),
    "Etc/GMT+4": ("SA Western Standard Time", "ZZ"),
    "Etc/GMT+5": ("SA Pacific Standard Time", "ZZ"),
    "Etc/GMT+6": ("Central America Standard Time", "ZZ"),
    "Etc/GMT+7": ("US Mountain Standard Time", "ZZ"),
    "Etc/GMT+8": ("UTC-08", "001"),
    "Etc/GMT+9": ("UTC-09", "001"),
    "Etc/GMT-1": ("W. Central Africa Standard Time", "ZZ"),
    "Etc/GMT-10": ("West Pacific Standard Time", "ZZ"),
    "Etc/GMT-11": ("Central Pacific Standard Time", "ZZ"),
    "Etc/GMT-12": ("UTC+12", "001"),
    "Etc/GMT-13": ("UTC+13", "001"),
    "Etc/GMT-14": ("Line Islands Standard Time", "ZZ"),
    "Etc/GMT-2": ("South Africa Standard Time", "ZZ"),
    "Etc/GMT-3": ("E. Africa Standard Time", "ZZ"),
    "Etc/GMT-4": ("Arabian Standard Time", "ZZ"),
    "Etc/GMT-5": ("West Asia Standard Time", "ZZ"),
    "Etc/GMT-6": ("Central Asia Standard Time", "ZZ"),
    "Etc/GMT-7": ("SE Asia Standard Time", "ZZ"),
    "Etc/GMT-8": ("Singapore Standard Time", "ZZ"),
    "Etc/GMT-9": ("Tokyo Standard Time", "ZZ"),
    "Etc/UTC": ("UTC", "001"),
    "Europe/Amsterdam": ("W. Europe Standard Time", "NL"),
    "Europe/Andorra": ("W. Europe Standard Time", "AD"),
    "Europe/Astrakhan": ("Astrakhan Standard Time", "001"),
    "Europe/Athens": ("GTB Standard Time", "GR"),
    "Europe/Belgrade": ("Central Europe Standard Time", "RS"),
    "Europe/Berlin": ("W. Europe Standard Time", "001"),
    "Europe/Bratislava": ("Central Europe Standard Time", "SK"),
    "Europe/Brussels": ("Romance Standard Time", "BE"),
    "Europe/Bucharest": ("GTB Standard Time", "001"),
    "Europe/Budapest": ("Central Europe Standard Time", "001"),
    "Europe/Busingen": ("W. Europe Standard Time", "DE"),
    "Europe/Chisinau": ("E. Europe Standard Time", "001"),
    "Europe/Copenhagen": ("Romance Standard Time", "DK"),
    "Europe/Dublin": ("GMT Standard Time", "IE"),
    "Europe/Gibraltar": ("W. Europe Standard Time", "GI"),
    "Europe/Guernsey": ("GMT Standard Time", "GG"),
    "Europe/Helsinki": ("FLE Standard Time", "FI"),
    "Europe/Isle_of_Man": ("GMT Standard Time", "IM"),
    "Europe/Istanbul": ("Turkey Standard Time", "001"),
    "Europe/Jersey": ("GMT Standard Time", "JE"),
    "Europe/Kaliningrad": ("Kaliningrad Standard Time", "001"),
    "Europe/Kiev": ("FLE Standard Time", "001"),
    "Europe/Kirov": ("Russian Standard Time", "RU"),
    "Europe/Lisbon": ("GMT Standard Time", "PT"),
    "Europe/Ljubljana": ("Central Europe Standard Time", "SI"),
    "Europe/London": ("GMT Standard Time", "001"),
    "Europe/Luxembourg": ("W. Europe Standard Time", "LU"),
    "Europe/Madrid": ("Romance Standard Time", "ES"),
    "Europe/Malta": ("W. Europe Standard Time", "MT"),
    "Europe/Mariehamn": ("FLE Standard Time", "AX"),
    "Europe/Minsk": ("Belarus Standard Time", "001"),
    "Europe/Monaco": ("W. Europe Standard Time", "MC"),
    "Europe/Moscow": ("Russian Standard Time", "001"),
    "Europe/Oslo": ("W. Europe Standard Time", "NO"),
    "Europe/Paris": ("Romance Standard Time", "001"),
    "Europe/Podgorica": ("Central Europe Standard Time", "ME"),
    "Europe/Prague": ("Central Europe Standard Time", "CZ"),
    "Europe/Riga": ("FLE Standard Time", "LV"),
    "Europe/Rome": ("W. Europe Standard Time", "IT"),
    "Europe/Samara": ("Russia Time Zone 3", "001"),
    "Europe/San_Marino": ("W. Europe Standard Time", "SM"),
    "Europe/Sarajevo": ("Central European Standard Time", "BA"),
    "Europe/Saratov": ("Saratov Standard Time", "001"),
    "Europe/Simferopol": ("Russian Standard Time", "UA"),
    "Europe/Skopje": ("Central European Standard Time", "MK"),
    "Europe/Sofia": ("FLE Standard Time", "BG"),
    "Europe/Stockholm": ("W. Europe Standard Time", "SE"),
    "Europe/Tallinn": ("FLE Standard Time", "EE"),
    "Europe/Tirane": ("Central Europe Standard Time", "AL"),
    "Europe/Ulyanovsk": ("Astrakhan Standard Time", "RU"),
    "Europe/Uzhgorod": ("FLE Standard Time", "UA"),
    "Europe/Vaduz": ("W. Europe Standard Time", "LI"),
    "Europe/Vatican": ("W. Europe Standard Time", "VA"),
    "Europe/Vienna": ("W. Europe Standard Time", "AT"),
    "Europe/Vilnius": ("FLE Standard Time", "LT"),
    "Europe/Volgograd": ("Volgograd Standard Time", "001"),
    "Europe/Warsaw": ("Central European Standard Time", "001"),
    "Europe/Zagreb": ("Central European Standard Time", "HR"),
    "Europe/Zaporozhye": ("FLE Standard Time", "UA"),
    "Europe/Zurich": ("W. Europe Standard Time", "CH"),
    "Indian/Antananarivo": ("E. Africa Standard Time", "MG"),
    "Indian/Chagos": ("Central Asia Standard Time", "IO"),
    "Indian/Christmas": ("SE Asia Standard Time", "CX"),
    "Indian/Cocos": ("Myanmar Standard Time", "CC"),
    "Indian/Comoro": ("E. Africa Standard Time", "KM"),
    "Indian/Kerguelen": ("West Asia Standard Time", "TF"),
    "Indian/Mahe": ("Mauritius Standard Time", "SC"),
    "Indian/Maldives": ("West Asia Standard Time", "MV"),
    "Indian/Mauritius": ("Mauritius Standard Time", "001"),
    "Indian/Mayotte": ("E. Africa Standard Time", "YT"),
    "Indian/Reunion": ("Mauritius Standard Time", "RE"),
    "MST7MDT": ("Mountain Standard Time", "ZZ"),
    "PST8PDT": ("Pacific Standard Time", "ZZ"),
    "Pacific/Apia": ("Samoa Standard Time", "001"),
    "Pacific/Auckland": ("New Zealand Standard Time", "001"),
    "Pacific/Bougainville": ("Bougainville Standard Time", "001"),
    "Pacific/Chatham": ("Chatham Islands Standard Time", "001"),
    "Pacific/Easter": ("Easter Island Standard Time", "001"),
    "Pacific/Efate": ("Central Pacific Standard Time", "VU"),
    "Pacific/Enderbury": ("UTC+13", "KI"),
    "Pacific/Fakaofo": ("UTC+13", "TK"),
    "Pacific/Fiji": ("Fiji Standard Time", "001"),
    "Pacific/Funafuti": ("UTC+12", "TV"),
    "Pacific/Galapagos": ("Central America Standard Time", "EC"),
    "Pacific/Gambier": ("UTC-09", "PF"),
    "Pacific/Guadalcanal": ("Central Pacific Standard Time", "001"),
    "Pacific/Guam": ("West Pacific Standard Time", "GU"),
    "Pacific/Honolulu": ("Hawaiian Standard Time", "001"),
    "Pacific/Johnston": ("Hawaiian Standard Time", "UM"),
    "Pacific/Kiritimati": ("Line Islands Standard Time", "001"),
    "Pacific/Kosrae": ("Central Pacific Standard Time", "FM"),
    "Pacific/Kwajalein": ("UTC+12", "MH"),
    "Pacific/Majuro": ("UTC+12", "MH"),
    "Pacific/Marquesas": ("Marquesas Standard Time", "001"),
    "Pacific/Midway": ("UTC-11", "UM"),
    "Pacific/Nauru": ("UTC+12", "NR"),
    "Pacific/Niue": ("UTC-11", "NU"),
    "Pacific/Norfolk": ("Norfolk Standard Time", "001"),
    "Pacific/Noumea": ("Central Pacific Standard Time", "NC"),
    "Pacific/Pago_Pago": ("UTC-11", "AS"),
    "Pacific/Palau": ("Tokyo Standard Time", "PW"),
    "Pacific/Pitcairn": ("UTC-08", "PN"),
    "Pacific/Ponape": ("Central Pacific Standard Time", "FM"),
    "Pacific/Port_Moresby": ("West Pacific Standard Time", "001"),
    "Pacific/Rarotonga": ("Hawaiian Standard Time", "CK"),
    "Pacific/Saipan": ("West Pacific Standard Time", "MP"),
    "Pacific/Tahiti": ("Hawaiian Standard Time", "PF"),
    "Pacific/Tarawa": ("UTC+12", "KI"),
    "Pacific/Tongatapu": ("Tonga Standard Time", "001"),
    "Pacific/Truk": ("West Pacific Standard Time", "FM"),
    "Pacific/Wake": ("UTC+12", "UM"),
    "Pacific/Wallis": ("UTC+12", "WF"),
}

# Timezone names used by IANA but not mentioned in the CLDR. All of them have an alias in CLDR. This is essentially
# all timezone names that zoneinfo.available_timezones() contains but CLDR doesn't. Aliases were found
# at https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
IANA_TO_MS_TIMEZONE_MAP = dict(
    CLDR_TO_MS_TIMEZONE_MAP,
    **{
        "Africa/Asmara": CLDR_TO_MS_TIMEZONE_MAP["Africa/Nairobi"],
        "Africa/Timbuktu": CLDR_TO_MS_TIMEZONE_MAP["Africa/Abidjan"],
        "America/Argentina/Buenos_Aires": CLDR_TO_MS_TIMEZONE_MAP["America/Buenos_Aires"],
        "America/Argentina/Catamarca": CLDR_TO_MS_TIMEZONE_MAP["America/Catamarca"],
        "America/Argentina/ComodRivadavia": CLDR_TO_MS_TIMEZONE_MAP["America/Catamarca"],
        "America/Argentina/Cordoba": CLDR_TO_MS_TIMEZONE_MAP["America/Cordoba"],
        "America/Argentina/Jujuy": CLDR_TO_MS_TIMEZONE_MAP["America/Jujuy"],
        "America/Argentina/Mendoza": CLDR_TO_MS_TIMEZONE_MAP["America/Mendoza"],
        "America/Atikokan": CLDR_TO_MS_TIMEZONE_MAP["America/Coral_Harbour"],
        "America/Atka": CLDR_TO_MS_TIMEZONE_MAP["America/Adak"],
        "America/Ensenada": CLDR_TO_MS_TIMEZONE_MAP["America/Tijuana"],
        "America/Fort_Wayne": CLDR_TO_MS_TIMEZONE_MAP["America/Indianapolis"],
        "America/Indiana/Indianapolis": CLDR_TO_MS_TIMEZONE_MAP["America/Indianapolis"],
        "America/Kentucky/Louisville": CLDR_TO_MS_TIMEZONE_MAP["America/Louisville"],
        "America/Knox_IN": CLDR_TO_MS_TIMEZONE_MAP["America/Indiana/Knox"],
        "America/Nuuk": CLDR_TO_MS_TIMEZONE_MAP["America/Godthab"],
        "America/Porto_Acre": CLDR_TO_MS_TIMEZONE_MAP["America/Rio_Branco"],
        "America/Rosario": CLDR_TO_MS_TIMEZONE_MAP["America/Cordoba"],
        "America/Shiprock": CLDR_TO_MS_TIMEZONE_MAP["America/Denver"],
        "America/Virgin": CLDR_TO_MS_TIMEZONE_MAP["America/Port_of_Spain"],
        "Antarctica/South_Pole": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Auckland"],
        "Antarctica/Troll": CLDR_TO_MS_TIMEZONE_MAP["Europe/Oslo"],
        "Asia/Ashkhabad": CLDR_TO_MS_TIMEZONE_MAP["Asia/Ashgabat"],
        "Asia/Chongqing": CLDR_TO_MS_TIMEZONE_MAP["Asia/Shanghai"],
        "Asia/Chungking": CLDR_TO_MS_TIMEZONE_MAP["Asia/Shanghai"],
        "Asia/Dacca": CLDR_TO_MS_TIMEZONE_MAP["Asia/Dhaka"],
        "Asia/Harbin": CLDR_TO_MS_TIMEZONE_MAP["Asia/Shanghai"],
        "Asia/Ho_Chi_Minh": CLDR_TO_MS_TIMEZONE_MAP["Asia/Saigon"],
        "Asia/Istanbul": CLDR_TO_MS_TIMEZONE_MAP["Europe/Istanbul"],
        "Asia/Kashgar": CLDR_TO_MS_TIMEZONE_MAP["Asia/Urumqi"],
        "Asia/Kathmandu": CLDR_TO_MS_TIMEZONE_MAP["Asia/Katmandu"],
        "Asia/Kolkata": CLDR_TO_MS_TIMEZONE_MAP["Asia/Calcutta"],
        "Asia/Macao": CLDR_TO_MS_TIMEZONE_MAP["Asia/Macau"],
        "Asia/Tel_Aviv": CLDR_TO_MS_TIMEZONE_MAP["Asia/Jerusalem"],
        "Asia/Thimbu": CLDR_TO_MS_TIMEZONE_MAP["Asia/Thimphu"],
        "Asia/Ujung_Pandang": CLDR_TO_MS_TIMEZONE_MAP["Asia/Makassar"],
        "Asia/Ulan_Bator": CLDR_TO_MS_TIMEZONE_MAP["Asia/Ulaanbaatar"],
        "Asia/Yangon": CLDR_TO_MS_TIMEZONE_MAP["Asia/Rangoon"],
        "Atlantic/Faroe": CLDR_TO_MS_TIMEZONE_MAP["Atlantic/Faeroe"],
        "Atlantic/Jan_Mayen": CLDR_TO_MS_TIMEZONE_MAP["Europe/Oslo"],
        "Australia/ACT": CLDR_TO_MS_TIMEZONE_MAP["Australia/Sydney"],
        "Australia/Canberra": CLDR_TO_MS_TIMEZONE_MAP["Australia/Sydney"],
        "Australia/LHI": CLDR_TO_MS_TIMEZONE_MAP["Australia/Lord_Howe"],
        "Australia/NSW": CLDR_TO_MS_TIMEZONE_MAP["Australia/Sydney"],
        "Australia/North": CLDR_TO_MS_TIMEZONE_MAP["Australia/Darwin"],
        "Australia/Queensland": CLDR_TO_MS_TIMEZONE_MAP["Australia/Brisbane"],
        "Australia/South": CLDR_TO_MS_TIMEZONE_MAP["Australia/Adelaide"],
        "Australia/Tasmania": CLDR_TO_MS_TIMEZONE_MAP["Australia/Hobart"],
        "Australia/Victoria": CLDR_TO_MS_TIMEZONE_MAP["Australia/Melbourne"],
        "Australia/West": CLDR_TO_MS_TIMEZONE_MAP["Australia/Perth"],
        "Australia/Yancowinna": CLDR_TO_MS_TIMEZONE_MAP["Australia/Broken_Hill"],
        "Brazil/Acre": CLDR_TO_MS_TIMEZONE_MAP["America/Rio_Branco"],
        "Brazil/DeNoronha": CLDR_TO_MS_TIMEZONE_MAP["America/Noronha"],
        "Brazil/East": CLDR_TO_MS_TIMEZONE_MAP["America/Sao_Paulo"],
        "Brazil/West": CLDR_TO_MS_TIMEZONE_MAP["America/Manaus"],
        "CET": CLDR_TO_MS_TIMEZONE_MAP["Europe/Paris"],
        "Canada/Atlantic": CLDR_TO_MS_TIMEZONE_MAP["America/Halifax"],
        "Canada/Central": CLDR_TO_MS_TIMEZONE_MAP["America/Winnipeg"],
        "Canada/Eastern": CLDR_TO_MS_TIMEZONE_MAP["America/Toronto"],
        "Canada/Mountain": CLDR_TO_MS_TIMEZONE_MAP["America/Edmonton"],
        "Canada/Newfoundland": CLDR_TO_MS_TIMEZONE_MAP["America/St_Johns"],
        "Canada/Pacific": CLDR_TO_MS_TIMEZONE_MAP["America/Vancouver"],
        "Canada/Saskatchewan": CLDR_TO_MS_TIMEZONE_MAP["America/Regina"],
        "Canada/Yukon": CLDR_TO_MS_TIMEZONE_MAP["America/Whitehorse"],
        "Chile/Continental": CLDR_TO_MS_TIMEZONE_MAP["America/Santiago"],
        "Chile/EasterIsland": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Easter"],
        "Cuba": CLDR_TO_MS_TIMEZONE_MAP["America/Havana"],
        "EET": CLDR_TO_MS_TIMEZONE_MAP["Europe/Sofia"],
        "EST": CLDR_TO_MS_TIMEZONE_MAP["America/Cancun"],
        "Egypt": CLDR_TO_MS_TIMEZONE_MAP["Africa/Cairo"],
        "Eire": CLDR_TO_MS_TIMEZONE_MAP["Europe/Dublin"],
        "Etc/GMT+0": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "Etc/GMT-0": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "Etc/GMT0": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "Etc/Greenwich": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "Etc/UCT": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "Etc/Universal": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "Etc/Zulu": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "Europe/Belfast": CLDR_TO_MS_TIMEZONE_MAP["Europe/London"],
        "Europe/Nicosia": CLDR_TO_MS_TIMEZONE_MAP["Asia/Nicosia"],
        "Europe/Tiraspol": CLDR_TO_MS_TIMEZONE_MAP["Europe/Chisinau"],
        "Factory": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "GB": CLDR_TO_MS_TIMEZONE_MAP["Europe/London"],
        "GB-Eire": CLDR_TO_MS_TIMEZONE_MAP["Europe/London"],
        "GMT": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "GMT+0": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "GMT-0": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "GMT0": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "Greenwich": CLDR_TO_MS_TIMEZONE_MAP["Etc/GMT"],
        "HST": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Honolulu"],
        "Hongkong": CLDR_TO_MS_TIMEZONE_MAP["Asia/Hong_Kong"],
        "Iceland": CLDR_TO_MS_TIMEZONE_MAP["Atlantic/Reykjavik"],
        "Iran": CLDR_TO_MS_TIMEZONE_MAP["Asia/Tehran"],
        "Israel": CLDR_TO_MS_TIMEZONE_MAP["Asia/Jerusalem"],
        "Jamaica": CLDR_TO_MS_TIMEZONE_MAP["America/Jamaica"],
        "Japan": CLDR_TO_MS_TIMEZONE_MAP["Asia/Tokyo"],
        "Kwajalein": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Kwajalein"],
        "Libya": CLDR_TO_MS_TIMEZONE_MAP["Africa/Tripoli"],
        "MET": CLDR_TO_MS_TIMEZONE_MAP["Europe/Paris"],
        "MST": CLDR_TO_MS_TIMEZONE_MAP["America/Phoenix"],
        "Mexico/BajaNorte": CLDR_TO_MS_TIMEZONE_MAP["America/Tijuana"],
        "Mexico/BajaSur": CLDR_TO_MS_TIMEZONE_MAP["America/Mazatlan"],
        "Mexico/General": CLDR_TO_MS_TIMEZONE_MAP["America/Mexico_City"],
        "NZ": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Auckland"],
        "NZ-CHAT": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Chatham"],
        "Navajo": CLDR_TO_MS_TIMEZONE_MAP["America/Denver"],
        "PRC": CLDR_TO_MS_TIMEZONE_MAP["Asia/Shanghai"],
        "Pacific/Chuuk": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Truk"],
        "Pacific/Kanton": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Enderbury"],
        "Pacific/Pohnpei": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Ponape"],
        "Pacific/Samoa": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Pago_Pago"],
        "Pacific/Yap": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Truk"],
        "Poland": CLDR_TO_MS_TIMEZONE_MAP["Europe/Warsaw"],
        "Portugal": CLDR_TO_MS_TIMEZONE_MAP["Europe/Lisbon"],
        "ROC": CLDR_TO_MS_TIMEZONE_MAP["Asia/Taipei"],
        "ROK": CLDR_TO_MS_TIMEZONE_MAP["Asia/Seoul"],
        "Singapore": CLDR_TO_MS_TIMEZONE_MAP["Asia/Singapore"],
        "Turkey": CLDR_TO_MS_TIMEZONE_MAP["Europe/Istanbul"],
        "UCT": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "US/Alaska": CLDR_TO_MS_TIMEZONE_MAP["America/Anchorage"],
        "US/Aleutian": CLDR_TO_MS_TIMEZONE_MAP["America/Adak"],
        "US/Arizona": CLDR_TO_MS_TIMEZONE_MAP["America/Phoenix"],
        "US/Central": CLDR_TO_MS_TIMEZONE_MAP["America/Chicago"],
        "US/East-Indiana": CLDR_TO_MS_TIMEZONE_MAP["America/Indianapolis"],
        "US/Eastern": CLDR_TO_MS_TIMEZONE_MAP["America/New_York"],
        "US/Hawaii": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Honolulu"],
        "US/Indiana-Starke": CLDR_TO_MS_TIMEZONE_MAP["America/Indiana/Knox"],
        "US/Michigan": CLDR_TO_MS_TIMEZONE_MAP["America/Detroit"],
        "US/Mountain": CLDR_TO_MS_TIMEZONE_MAP["America/Denver"],
        "US/Pacific": CLDR_TO_MS_TIMEZONE_MAP["America/Los_Angeles"],
        "US/Samoa": CLDR_TO_MS_TIMEZONE_MAP["Pacific/Pago_Pago"],
        "UTC": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "Universal": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
        "W-SU": CLDR_TO_MS_TIMEZONE_MAP["Europe/Moscow"],
        "WET": CLDR_TO_MS_TIMEZONE_MAP["Europe/Lisbon"],
        "Zulu": CLDR_TO_MS_TIMEZONE_MAP["Etc/UTC"],
    },
)

# Reverse map from Microsoft timezone ID to IANA timezone name. Non-IANA timezone ID's can be added here.
MS_TIMEZONE_TO_IANA_MAP = dict(
    # Use the CLDR map because the IANA map contains deprecated aliases that not all systems support
    {v[0]: k for k, v in CLDR_TO_MS_TIMEZONE_MAP.items() if v[1] == DEFAULT_TERRITORY},
    **{
        "tzone://Microsoft/Utc": "UTC",
    },
)



import logging

from ..properties import Notification
from ..util import create_element
from .common import EWSAccountService, add_xml_child

log = logging.getLogger(__name__)


class GetEvents(EWSAccountService):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getevents-operation
    """

    SERVICE_NAME = "GetEvents"
    prefer_affinity = True

    def call(self, subscription_id, watermark):
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(
                    subscription_id=subscription_id,
                    watermark=watermark,
                )
            )
        )

    def _elem_to_obj(self, elem):
        return Notification.from_xml(elem=elem, account=None)

    @classmethod
    def _get_elements_in_container(cls, container):
        return container.findall(Notification.response_tag())

    def get_payload(self, subscription_id, watermark):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        add_xml_child(payload, "m:SubscriptionId", subscription_id)
        add_xml_child(payload, "m:Watermark", watermark)
        return payload



from ..errors import InvalidTypeError
from ..folders import BaseFolder
from ..items import Item
from ..properties import FolderId
from ..util import MNS, create_element
from .common import EWSAccountService, folder_ids_element, item_ids_element


class MoveItem(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/moveitem-operation"""

    SERVICE_NAME = "MoveItem"
    element_container_name = f"{{{MNS}}}Items"

    def call(self, items, to_folder):
        if not isinstance(to_folder, (BaseFolder, FolderId)):
            raise InvalidTypeError("to_folder", to_folder, (BaseFolder, FolderId))
        return self._elems_to_objs(self._chunked_get_elements(self.get_payload, items=items, to_folder=to_folder))

    def _elem_to_obj(self, elem):
        return Item.id_from_xml(elem)

    def get_payload(self, items, to_folder):
        # Takes a list of items and returns their new item IDs
        payload = create_element(f"m:{self.SERVICE_NAME}")
        payload.append(folder_ids_element(folders=[to_folder], version=self.account.version, tag="m:ToFolderId"))
        payload.append(item_ids_element(items=items, version=self.account.version))
        return payload



import datetime
import logging
from decimal import Decimal

from ..ewsdatetime import UTC, EWSDateTime
from ..fields import (
    BooleanField,
    CharField,
    Choice,
    ChoiceField,
    DateTimeBackedDateField,
    DateTimeField,
    DecimalField,
    IntegerField,
    TaskRecurrenceField,
    TextField,
    TextListField,
)
from .item import Item

log = logging.getLogger(__name__)


class Task(Item):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/task"""

    ELEMENT_NAME = "Task"
    NOT_STARTED = "NotStarted"
    COMPLETED = "Completed"

    actual_work = IntegerField(field_uri="task:ActualWork", min=0)
    assigned_time = DateTimeField(field_uri="task:AssignedTime", is_read_only=True)
    billing_information = TextField(field_uri="task:BillingInformation")
    change_count = IntegerField(field_uri="task:ChangeCount", is_read_only=True, min=0)
    companies = TextListField(field_uri="task:Companies")
    # 'complete_date' can be set, but is ignored by the server, which sets it to now()
    complete_date = DateTimeField(field_uri="task:CompleteDate", is_read_only=True)
    contacts = TextListField(field_uri="task:Contacts")
    delegation_state = ChoiceField(
        field_uri="task:DelegationState",
        choices={
            Choice("NoMatch"),
            Choice("OwnNew"),
            Choice("Owned"),
            Choice("Accepted"),
            Choice("Declined"),
            Choice("Max"),
        },
        is_read_only=True,
    )
    delegator = CharField(field_uri="task:Delegator", is_read_only=True)
    due_date = DateTimeBackedDateField(field_uri="task:DueDate")
    is_editable = BooleanField(field_uri="task:IsAssignmentEditable", is_read_only=True)
    is_complete = BooleanField(field_uri="task:IsComplete", is_read_only=True)
    is_recurring = BooleanField(field_uri="task:IsRecurring", is_read_only=True)
    is_team_task = BooleanField(field_uri="task:IsTeamTask", is_read_only=True)
    mileage = TextField(field_uri="task:Mileage")
    owner = CharField(field_uri="task:Owner", is_read_only=True)
    percent_complete = DecimalField(
        field_uri="task:PercentComplete",
        is_required=True,
        default=Decimal(0.0),
        min=Decimal(0),
        max=Decimal(100),
        is_searchable=False,
    )
    recurrence = TaskRecurrenceField(field_uri="task:Recurrence", is_searchable=False)
    start_date = DateTimeBackedDateField(field_uri="task:StartDate")
    status = ChoiceField(
        field_uri="task:Status",
        choices={
            Choice(NOT_STARTED),
            Choice("InProgress"),
            Choice(COMPLETED),
            Choice("WaitingOnOthers"),
            Choice("Deferred"),
        },
        is_required=True,
        is_searchable=False,
        default=NOT_STARTED,
    )
    status_description = CharField(field_uri="task:StatusDescription", is_read_only=True)
    total_work = IntegerField(field_uri="task:TotalWork", min=0)

    def clean(self, version=None):
        super().clean(version=version)
        if self.due_date and self.start_date and self.due_date < self.start_date:
            log.warning(
                "'due_date' must be greater than 'start_date' (%s vs %s). Resetting 'due_date'",
                self.due_date,
                self.start_date,
            )
            self.due_date = self.start_date
        if self.complete_date:
            if self.status != self.COMPLETED:
                log.warning(
                    "'status' must be '%s' when 'complete_date' is set (%s). Resetting", self.COMPLETED, self.status
                )
                self.status = self.COMPLETED
            now = datetime.datetime.now(tz=UTC)
            if (self.complete_date - now).total_seconds() > 120:
                # Reset complete_date values that are in the future
                # 'complete_date' can be set automatically by the server. Allow some grace between local and server time
                log.warning("'complete_date' must be in the past (%s vs %s). Resetting", self.complete_date, now)
                self.complete_date = now
            if self.start_date and self.complete_date.date() < self.start_date:
                log.warning(
                    "'complete_date' must be greater than 'start_date' (%s vs %s). Resetting",
                    self.complete_date,
                    self.start_date,
                )
                self.complete_date = EWSDateTime.combine(self.start_date, datetime.time(0, 0)).replace(tzinfo=UTC)
        if self.percent_complete is not None:
            if self.status == self.COMPLETED and self.percent_complete != Decimal(100):
                # percent_complete must be 100% if task is complete
                log.warning(
                    "'percent_complete' must be 100 when 'status' is '%s' (%s). Resetting",
                    self.COMPLETED,
                    self.percent_complete,
                )
                self.percent_complete = Decimal(100)
            elif self.status == self.NOT_STARTED and self.percent_complete != Decimal(0):
                # percent_complete must be 0% if task is not started
                log.warning(
                    "'percent_complete' must be 0 when 'status' is '%s' (%s). Resetting",
                    self.NOT_STARTED,
                    self.percent_complete,
                )
                self.percent_complete = Decimal(0)

    def complete(self):
        # A helper method to mark a task as complete on the server
        self.status = Task.COMPLETED
        self.percent_complete = Decimal(100)
        self.save()



from ..errors import InvalidEnumValue
from ..folders import Folder
from ..folders.queryset import FOLDER_TRAVERSAL_CHOICES
from ..items import SHAPE_CHOICES
from ..util import MNS, TNS, create_element
from ..version import EXCHANGE_2010
from .common import EWSPagingService, folder_ids_element, shape_element


class FindFolder(EWSPagingService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/findfolder-operation"""

    SERVICE_NAME = "FindFolder"
    element_container_name = f"{{{TNS}}}Folders"
    paging_container_name = f"{{{MNS}}}RootFolder"
    supports_paging = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.root = None  # A hack to communicate parsing args to _elems_to_objs()

    def call(self, folders, additional_fields, restriction, shape, depth, max_items, offset):
        """Find subfolders of a folder.

        :param folders: the folders to act on
        :param additional_fields: the extra fields that should be returned with the folder, as FieldPath objects
        :param restriction: Restriction object that defines the filters for the query
        :param shape: The set of attributes to return
        :param depth: How deep in the folder structure to search for folders
        :param max_items: The maximum number of items to return
        :param offset: the offset relative to the first item in the item collection. Usually 0.

        :return: XML elements for the matching folders
        """
        if shape not in SHAPE_CHOICES:
            raise InvalidEnumValue("shape", shape, SHAPE_CHOICES)
        if depth not in FOLDER_TRAVERSAL_CHOICES:
            raise InvalidEnumValue("depth", depth, FOLDER_TRAVERSAL_CHOICES)
        roots = {f.root for f in folders}
        if len(roots) != 1:
            raise ValueError(f"All folders in 'roots' must have the same root hierarchy ({roots})")
        self.root = roots.pop()
        return self._elems_to_objs(
            self._paged_call(
                payload_func=self.get_payload,
                max_items=max_items,
                folders=folders,
                **dict(
                    additional_fields=additional_fields,
                    restriction=restriction,
                    shape=shape,
                    depth=depth,
                    page_size=self.page_size,
                    offset=offset,
                ),
            )
        )

    def _elem_to_obj(self, elem):
        return Folder.from_xml_with_root(elem=elem, root=self.root)

    def get_payload(self, folders, additional_fields, restriction, shape, depth, page_size, offset=0):
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=dict(Traversal=depth))
        payload.append(
            shape_element(
                tag="m:FolderShape", shape=shape, additional_fields=additional_fields, version=self.account.version
            )
        )
        if self.account.version.build >= EXCHANGE_2010:
            indexed_page_folder_view = create_element(
                "m:IndexedPageFolderView",
                attrs=dict(MaxEntriesReturned=page_size, Offset=offset, BasePoint="Beginning"),
            )
            payload.append(indexed_page_folder_view)
        else:
            if offset != 0:
                raise NotImplementedError("'offset' is only supported for Exchange 2010 servers and later")
        if restriction:
            payload.append(restriction.to_xml(version=self.account.version))
        payload.append(folder_ids_element(folders=folders, version=self.protocol.version, tag="m:ParentFolderIds"))
        return payload



from ..errors import InvalidEnumValue
from ..items import AFFECTED_TASK_OCCURRENCES_CHOICES, DELETE_TYPE_CHOICES, SEND_MEETING_CANCELLATIONS_CHOICES
from ..util import create_element
from ..version import EXCHANGE_2013_SP1
from .common import EWSAccountService, item_ids_element


class DeleteItem(EWSAccountService):
    """Take a folder and a list of (id, changekey) tuples. Return result of deletion as a list of tuples
    (success[True|False], errormessage), in the same order as the input list.

    MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/deleteitem-operation
    """

    SERVICE_NAME = "DeleteItem"
    returns_elements = False

    def call(self, items, delete_type, send_meeting_cancellations, affected_task_occurrences, suppress_read_receipts):
        if delete_type not in DELETE_TYPE_CHOICES:
            raise InvalidEnumValue("delete_type", delete_type, DELETE_TYPE_CHOICES)
        if send_meeting_cancellations not in SEND_MEETING_CANCELLATIONS_CHOICES:
            raise InvalidEnumValue(
                "send_meeting_cancellations", send_meeting_cancellations, SEND_MEETING_CANCELLATIONS_CHOICES
            )
        if affected_task_occurrences not in AFFECTED_TASK_OCCURRENCES_CHOICES:
            raise InvalidEnumValue(
                "affected_task_occurrences", affected_task_occurrences, AFFECTED_TASK_OCCURRENCES_CHOICES
            )
        return self._chunked_get_elements(
            self.get_payload,
            items=items,
            delete_type=delete_type,
            send_meeting_cancellations=send_meeting_cancellations,
            affected_task_occurrences=affected_task_occurrences,
            suppress_read_receipts=suppress_read_receipts,
        )

    def get_payload(
        self, items, delete_type, send_meeting_cancellations, affected_task_occurrences, suppress_read_receipts
    ):
        # Takes a list of (id, changekey) tuples or Item objects and returns the XML for a DeleteItem request.
        attrs = dict(
            DeleteType=delete_type,
            SendMeetingCancellations=send_meeting_cancellations,
            AffectedTaskOccurrences=affected_task_occurrences,
        )
        if self.account.version.build >= EXCHANGE_2013_SP1:
            attrs["SuppressReadReceipts"] = suppress_read_receipts
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=attrs)
        payload.append(item_ids_element(items=items, version=self.account.version))
        return payload



from ..util import create_element, set_xml_value
from .common import EWSAccountService


class DeleteUserConfiguration(EWSAccountService):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/deleteuserconfiguration-operation
    """

    SERVICE_NAME = "DeleteUserConfiguration"
    returns_elements = False

    def call(self, user_configuration_name):
        return self._get_elements(payload=self.get_payload(user_configuration_name=user_configuration_name))

    def get_payload(self, user_configuration_name):
        return set_xml_value(
            create_element(f"m:{self.SERVICE_NAME}"), user_configuration_name, version=self.account.version
        )



from .account import Account, Identity
from .attachments import FileAttachment, ItemAttachment
from .autodiscover import discover
from .configuration import Configuration
from .credentials import DELEGATE, IMPERSONATION, Credentials, OAuth2AuthorizationCodeCredentials, OAuth2Credentials
from .ewsdatetime import UTC, UTC_NOW, EWSDate, EWSDateTime, EWSTimeZone
from .extended_properties import ExtendedProperty
from .folders import DEEP, SHALLOW, Folder, FolderCollection, RootOfHierarchy
from .items import (
    AcceptItem,
    CalendarItem,
    CancelCalendarItem,
    Contact,
    DeclineItem,
    DistributionList,
    ForwardItem,
    Message,
    PostItem,
    ReplyAllToItem,
    ReplyToItem,
    Task,
    TentativelyAcceptItem,
)
from .properties import UID, Attendee, Body, DLMailbox, HTMLBody, ItemId, Mailbox, Room, RoomList
from .protocol import BaseProtocol, FailFast, FaultTolerance, NoVerifyHTTPAdapter, TLSClientAuth
from .restriction import Q
from .settings import OofSettings
from .transport import BASIC, CBA, DIGEST, GSSAPI, NTLM, OAUTH2, SSPI
from .version import Build, Version

__version__ = "4.7.3"

__all__ = [
    "__version__",
    "Account",
    "Identity",
    "FileAttachment",
    "ItemAttachment",
    "discover",
    "Configuration",
    "DELEGATE",
    "IMPERSONATION",
    "Credentials",
    "OAuth2AuthorizationCodeCredentials",
    "OAuth2Credentials",
    "EWSDate",
    "EWSDateTime",
    "EWSTimeZone",
    "UTC",
    "UTC_NOW",
    "ExtendedProperty",
    "Folder",
    "RootOfHierarchy",
    "FolderCollection",
    "SHALLOW",
    "DEEP",
    "AcceptItem",
    "TentativelyAcceptItem",
    "DeclineItem",
    "CalendarItem",
    "CancelCalendarItem",
    "Contact",
    "DistributionList",
    "Message",
    "PostItem",
    "Task",
    "ForwardItem",
    "ReplyToItem",
    "ReplyAllToItem",
    "ItemId",
    "Mailbox",
    "DLMailbox",
    "Attendee",
    "Room",
    "RoomList",
    "Body",
    "HTMLBody",
    "UID",
    "FailFast",
    "FaultTolerance",
    "BaseProtocol",
    "NoVerifyHTTPAdapter",
    "TLSClientAuth",
    "OofSettings",
    "Q",
    "BASIC",
    "DIGEST",
    "NTLM",
    "GSSAPI",
    "SSPI",
    "OAUTH2",
    "CBA",
    "Build",
    "Version",
    "close_connections",
]

# Set a default user agent, e.g. "exchangelib/3.1.1 (python-requests/2.22.0)"
import requests.utils

BaseProtocol.USERAGENT = f"{__name__}/{__version__} ({requests.utils.default_user_agent()})"


def close_connections():
    from .autodiscover import close_connections as close_autodiscover_connections
    from .protocol import close_connections as close_protocol_connections

    close_autodiscover_connections()
    close_protocol_connections()



import abc
import logging

from ..properties import FolderId
from ..util import MNS, TNS, create_element, xml_text_to_value
from .common import EWSPagingService, add_xml_child, folder_ids_element, parse_folder_elem, shape_element

log = logging.getLogger(__name__)


class SyncFolder(EWSPagingService, metaclass=abc.ABCMeta):
    """Base class for SyncFolderHierarchy and SyncFolderItems."""

    element_container_name = f"{{{MNS}}}Changes"
    # Change types
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    CHANGE_TYPES = (CREATE, UPDATE, DELETE)
    shape_tag = None
    last_in_range_name = None
    change_types_map = {
        f"{{{TNS}}}Create": CREATE,
        f"{{{TNS}}}Update": UPDATE,
        f"{{{TNS}}}Delete": DELETE,
    }

    def __init__(self, *args, **kwargs):
        # These values are reset and set each time call() is consumed
        self.sync_state = None
        self.includes_last_item_in_range = None
        super().__init__(*args, **kwargs)

    def _get_element_container(self, message, name=None):
        self.sync_state = message.find(f"{{{MNS}}}SyncState").text
        log.debug("Sync state is: %s", self.sync_state)
        self.includes_last_item_in_range = xml_text_to_value(message.find(self.last_in_range_name).text, bool)
        log.debug("Includes last item in range: %s", self.includes_last_item_in_range)
        return super()._get_element_container(message=message, name=name)

    def _partial_get_payload(self, folder, shape, additional_fields, sync_state):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        payload.append(
            shape_element(
                tag=self.shape_tag, shape=shape, additional_fields=additional_fields, version=self.account.version
            )
        )
        payload.append(folder_ids_element(folders=[folder], version=self.account.version, tag="m:SyncFolderId"))
        if sync_state:
            add_xml_child(payload, "m:SyncState", sync_state)
        return payload


class SyncFolderHierarchy(SyncFolder):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/syncfolderhierarchy-operation
    """

    SERVICE_NAME = "SyncFolderHierarchy"
    shape_tag = "m:FolderShape"
    last_in_range_name = f"{{{MNS}}}IncludesLastFolderInRange"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.folder = None  # A hack to communicate parsing args to _elems_to_objs()

    def call(self, folder, shape, additional_fields, sync_state):
        self.sync_state = sync_state
        self.folder = folder
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(
                    folder=folder,
                    shape=shape,
                    additional_fields=additional_fields,
                    sync_state=sync_state,
                )
            )
        )

    def _elem_to_obj(self, elem):
        change_type = self.change_types_map[elem.tag]
        if change_type == self.DELETE:
            folder = FolderId.from_xml(elem=elem.find(FolderId.response_tag()), account=self.account)
        else:
            # We can't find() the element because we don't know which tag to look for. The change element can
            # contain multiple folder types, each with their own tag.
            folder_elem = elem[0]
            folder = parse_folder_elem(elem=folder_elem, folder=self.folder, account=self.account)
        return change_type, folder

    def get_payload(self, folder, shape, additional_fields, sync_state):
        return self._partial_get_payload(
            folder=folder, shape=shape, additional_fields=additional_fields, sync_state=sync_state
        )



from ..properties import DelegateUser, DLMailbox, UserId  # The service expects a Mailbox element in the MNS namespace
from ..util import MNS, create_element, set_xml_value
from ..version import EXCHANGE_2007_SP1
from .common import EWSAccountService


class GetDelegate(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getdelegate-operation"""

    SERVICE_NAME = "GetDelegate"
    ERRORS_TO_CATCH_IN_RESPONSE = ()
    supported_from = EXCHANGE_2007_SP1

    def call(self, user_ids, include_permissions):
        return self._elems_to_objs(
            self._chunked_get_elements(
                self.get_payload,
                items=user_ids or [None],
                mailbox=DLMailbox(email_address=self.account.primary_smtp_address),
                include_permissions=include_permissions,
            )
        )

    def _elem_to_obj(self, elem):
        return DelegateUser.from_xml(elem=elem, account=self.account)

    def get_payload(self, user_ids, mailbox, include_permissions):
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=dict(IncludePermissions=include_permissions))
        set_xml_value(payload, mailbox, version=self.protocol.version)
        if user_ids != [None]:
            user_ids_elem = create_element("m:UserIds")
            for user_id in user_ids:
                if isinstance(user_id, str):
                    user_id = UserId(primary_smtp_address=user_id)
                set_xml_value(user_ids_elem, user_id, version=self.protocol.version)
            set_xml_value(payload, user_ids_elem, version=self.protocol.version)
        return payload

    @classmethod
    def _get_elements_in_container(cls, container):
        return container.findall(DelegateUser.response_tag())

    @classmethod
    def _response_message_tag(cls):
        return f"{{{MNS}}}DelegateUserResponseMessageType"



"""Implement a selection of EWS services (operations).

Exchange is very picky about things like the order of XML elements in SOAP requests, so we need to generate XML
automatically instead of taking advantage of Python SOAP libraries and the WSDL file.

Exchange EWS operations overview:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/ews-operations-in-exchange
"""

from .archive_item import ArchiveItem
from .common import CHUNK_SIZE, PAGE_SIZE
from .convert_id import ConvertId
from .copy_item import CopyItem
from .create_attachment import CreateAttachment
from .create_folder import CreateFolder
from .create_item import CreateItem
from .create_user_configuration import CreateUserConfiguration
from .delete_attachment import DeleteAttachment
from .delete_folder import DeleteFolder
from .delete_item import DeleteItem
from .delete_user_configuration import DeleteUserConfiguration
from .empty_folder import EmptyFolder
from .expand_dl import ExpandDL
from .export_items import ExportItems
from .find_folder import FindFolder
from .find_item import FindItem
from .find_people import FindPeople
from .get_attachment import GetAttachment
from .get_delegate import GetDelegate
from .get_events import GetEvents
from .get_folder import GetFolder
from .get_item import GetItem
from .get_mail_tips import GetMailTips
from .get_persona import GetPersona
from .get_room_lists import GetRoomLists
from .get_rooms import GetRooms
from .get_searchable_mailboxes import GetSearchableMailboxes
from .get_server_time_zones import GetServerTimeZones
from .get_streaming_events import GetStreamingEvents
from .get_user_availability import GetUserAvailability
from .get_user_configuration import GetUserConfiguration
from .get_user_oof_settings import GetUserOofSettings
from .mark_as_junk import MarkAsJunk
from .move_folder import MoveFolder
from .move_item import MoveItem
from .resolve_names import ResolveNames
from .send_item import SendItem
from .send_notification import SendNotification
from .set_user_oof_settings import SetUserOofSettings
from .subscribe import SubscribeToPull, SubscribeToPush, SubscribeToStreaming
from .sync_folder_hierarchy import SyncFolderHierarchy
from .sync_folder_items import SyncFolderItems
from .unsubscribe import Unsubscribe
from .update_folder import UpdateFolder
from .update_item import UpdateItem
from .update_user_configuration import UpdateUserConfiguration
from .upload_items import UploadItems

__all__ = [
    "CHUNK_SIZE",
    "PAGE_SIZE",
    "ArchiveItem",
    "ConvertId",
    "CopyItem",
    "CreateAttachment",
    "CreateFolder",
    "CreateItem",
    "CreateUserConfiguration",
    "DeleteAttachment",
    "DeleteFolder",
    "DeleteUserConfiguration",
    "DeleteItem",
    "EmptyFolder",
    "ExpandDL",
    "ExportItems",
    "FindFolder",
    "FindItem",
    "FindPeople",
    "GetAttachment",
    "GetDelegate",
    "GetEvents",
    "GetFolder",
    "GetItem",
    "GetMailTips",
    "GetPersona",
    "GetRoomLists",
    "GetRooms",
    "GetSearchableMailboxes",
    "GetServerTimeZones",
    "GetStreamingEvents",
    "GetUserAvailability",
    "GetUserConfiguration",
    "GetUserOofSettings",
    "MarkAsJunk",
    "MoveFolder",
    "MoveItem",
    "ResolveNames",
    "SendItem",
    "SendNotification",
    "SetUserOofSettings",
    "SubscribeToPull",
    "SubscribeToPush",
    "SubscribeToStreaming",
    "SyncFolderHierarchy",
    "SyncFolderItems",
    "Unsubscribe",
    "UpdateFolder",
    "UpdateItem",
    "UpdateUserConfiguration",
    "UploadItems",
]



from ..properties import ItemId, ParentFolderId
from ..util import MNS, add_xml_child, create_element, set_xml_value
from .common import EWSAccountService, to_item_id


class UploadItems(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/uploaditems-operation"""

    SERVICE_NAME = "UploadItems"
    element_container_name = f"{{{MNS}}}ItemId"

    def call(self, items):
        # _pool_requests expects 'items', not 'data'
        return self._elems_to_objs(self._chunked_get_elements(self.get_payload, items=items))

    def get_payload(self, items):
        """Upload given items to given account.

        'items' is an iterable of tuples where the first element is a Folder instance representing the ParentFolder
        that the item will be placed in and the second element is a tuple containing an optional ItemId, an optional
        Item.is_associated boolean, and a Data string returned from an ExportItems.
        call.

        :param items:
        """
        payload = create_element(f"m:{self.SERVICE_NAME}")
        items_elem = create_element("m:Items")
        payload.append(items_elem)
        for parent_folder, (item_id, is_associated, data_str) in items:
            # TODO: The full spec also allows the "UpdateOrCreate" create action.
            attrs = dict(CreateAction="Update" if item_id else "CreateNew")
            if is_associated is not None:
                attrs["IsAssociated"] = is_associated
            item = create_element("t:Item", attrs=attrs)
            set_xml_value(item, ParentFolderId(parent_folder.id, parent_folder.changekey), version=self.account.version)
            if item_id:
                set_xml_value(item, to_item_id(item_id, ItemId), version=self.account.version)
            add_xml_child(item, "t:Data", data_str)
            items_elem.append(item)
        return payload

    def _elem_to_obj(self, elem):
        return elem.get(ItemId.ID_ATTR), elem.get(ItemId.CHANGEKEY_ATTR)

    @classmethod
    def _get_elements_in_container(cls, container):
        return [container]



from .cache import AutodiscoverCache, autodiscover_cache
from .discovery import Autodiscovery, discover
from .protocol import AutodiscoverProtocol


def close_connections():
    with autodiscover_cache:
        autodiscover_cache.close()


def clear_cache():
    with autodiscover_cache:
        autodiscover_cache.clear()


__all__ = [
    "AutodiscoverCache",
    "AutodiscoverProtocol",
    "Autodiscovery",
    "discover",
    "autodiscover_cache",
    "close_connections",
    "clear_cache",
]



"""The 'Subscribe' service has two different modes, pull and push, with different signatures. Implement as two distinct
classes.
"""
import abc

from ..util import MNS, create_element
from .common import EWSAccountService, add_xml_child, folder_ids_element


class Subscribe(EWSAccountService, metaclass=abc.ABCMeta):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/subscribe-operation"""

    SERVICE_NAME = "Subscribe"
    EVENT_TYPES = (
        "CopiedEvent",
        "CreatedEvent",
        "DeletedEvent",
        "ModifiedEvent",
        "MovedEvent",
        "NewMailEvent",
        "FreeBusyChangedEvent",
    )
    subscription_request_elem_tag = None

    def _partial_call(self, payload_func, folders, event_types, **kwargs):
        if set(event_types) - set(self.EVENT_TYPES):
            raise ValueError(f"'event_types' values must consist of values in {self.EVENT_TYPES}")
        return self._elems_to_objs(
            self._get_elements(payload=payload_func(folders=folders, event_types=event_types, **kwargs))
        )

    def _elem_to_obj(self, elem):
        subscription_elem, watermark_elem = elem
        return subscription_elem.text, watermark_elem.text

    @classmethod
    def _get_elements_in_container(cls, container):
        return [(container.find(f"{{{MNS}}}SubscriptionId"), container.find(f"{{{MNS}}}Watermark"))]

    def _partial_payload(self, folders, event_types):
        request_elem = create_element(self.subscription_request_elem_tag)
        folder_ids = folder_ids_element(folders=folders, version=self.account.version, tag="t:FolderIds")
        request_elem.append(folder_ids)
        event_types_elem = create_element("t:EventTypes")
        for event_type in event_types:
            add_xml_child(event_types_elem, "t:EventType", event_type)
        if not len(event_types_elem):
            raise ValueError("'event_types' must not be empty")
        request_elem.append(event_types_elem)
        return request_elem


class SubscribeToPull(Subscribe):
    subscription_request_elem_tag = "m:PullSubscriptionRequest"
    prefer_affinity = True

    def call(self, folders, event_types, watermark, timeout):
        yield from self._partial_call(
            payload_func=self.get_payload,
            folders=folders,
            event_types=event_types,
            timeout=timeout,
            watermark=watermark,
        )

    def get_payload(self, folders, event_types, watermark, timeout):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        request_elem = self._partial_payload(folders=folders, event_types=event_types)
        if watermark:
            add_xml_child(request_elem, "m:Watermark", watermark)
        add_xml_child(request_elem, "t:Timeout", timeout)  # In minutes
        payload.append(request_elem)
        return payload


class SubscribeToPush(Subscribe):
    subscription_request_elem_tag = "m:PushSubscriptionRequest"

    def call(self, folders, event_types, watermark, status_frequency, url):
        yield from self._partial_call(
            payload_func=self.get_payload,
            folders=folders,
            event_types=event_types,
            status_frequency=status_frequency,
            url=url,
            watermark=watermark,
        )

    def get_payload(self, folders, event_types, watermark, status_frequency, url):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        request_elem = self._partial_payload(folders=folders, event_types=event_types)
        if watermark:
            add_xml_child(request_elem, "m:Watermark", watermark)
        add_xml_child(request_elem, "t:StatusFrequency", status_frequency)  # In minutes
        add_xml_child(request_elem, "t:URL", url)
        payload.append(request_elem)
        return payload


class SubscribeToStreaming(Subscribe):
    subscription_request_elem_tag = "m:StreamingSubscriptionRequest"
    prefer_affinity = True

    def call(self, folders, event_types):
        yield from self._partial_call(payload_func=self.get_payload, folders=folders, event_types=event_types)

    def _elem_to_obj(self, elem):
        return elem.text

    @classmethod
    def _get_elements_in_container(cls, container):
        return [container.find(f"{{{MNS}}}SubscriptionId")]

    def get_payload(self, folders, event_types):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        payload.append(self._partial_payload(folders=folders, event_types=event_types))
        return payload



import logging

from .fields import Choice, EmailSubField, LabelField, NamedSubField, SubField
from .properties import EWSElement, EWSMeta

log = logging.getLogger(__name__)


class IndexedElement(EWSElement, metaclass=EWSMeta):
    """Base class for all classes that implement an indexed element."""

    LABEL_CHOICES = ()


class SingleFieldIndexedElement(IndexedElement, metaclass=EWSMeta):
    """Base class for all classes that implement an indexed element with a single field."""

    @classmethod
    def value_field(cls, version):
        fields = cls.supported_fields(version=version)
        if len(fields) != 1:
            raise ValueError(f"Class {cls} must have only one value field (found {tuple(f.name for f in fields)})")
        return fields[0]


class EmailAddress(SingleFieldIndexedElement):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/entry-emailaddress"""

    ELEMENT_NAME = "Entry"
    LABEL_CHOICES = ("EmailAddress1", "EmailAddress2", "EmailAddress3")

    label = LabelField(field_uri="Key", choices={Choice(c) for c in LABEL_CHOICES}, default=LABEL_CHOICES[0])
    email = EmailSubField(is_required=True)


class PhoneNumber(SingleFieldIndexedElement):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/entry-phonenumber"""

    ELEMENT_NAME = "Entry"
    LABEL_CHOICES = (
        "AssistantPhone",
        "BusinessFax",
        "BusinessPhone",
        "BusinessPhone2",
        "Callback",
        "CarPhone",
        "CompanyMainPhone",
        "HomeFax",
        "HomePhone",
        "HomePhone2",
        "Isdn",
        "MobilePhone",
        "OtherFax",
        "OtherTelephone",
        "Pager",
        "PrimaryPhone",
        "RadioPhone",
        "Telex",
        "TtyTddPhone",
    )

    label = LabelField(field_uri="Key", choices={Choice(c) for c in LABEL_CHOICES}, default="PrimaryPhone")
    phone_number = SubField(is_required=True)


class MultiFieldIndexedElement(IndexedElement, metaclass=EWSMeta):
    """Base class for all classes that implement an indexed element with multiple fields."""


class PhysicalAddress(MultiFieldIndexedElement):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/entry-physicaladdress"""

    ELEMENT_NAME = "Entry"
    LABEL_CHOICES = ("Business", "Home", "Other")

    label = LabelField(field_uri="Key", choices={Choice(c) for c in LABEL_CHOICES}, default=LABEL_CHOICES[0])
    street = NamedSubField(field_uri="Street")  # Street, house number, etc.
    city = NamedSubField(field_uri="City")
    state = NamedSubField(field_uri="State")
    country = NamedSubField(field_uri="CountryOrRegion")
    zipcode = NamedSubField(field_uri="PostalCode")

    def clean(self, version=None):
        if isinstance(self.zipcode, int):
            self.zipcode = str(self.zipcode)
        super().clean(version=version)



import logging

from ..errors import ErrorNameResolutionMultipleResults, ErrorNameResolutionNoResults, InvalidEnumValue
from ..items import SEARCH_SCOPE_CHOICES, SHAPE_CHOICES, Contact
from ..properties import Mailbox
from ..util import MNS, add_xml_child, create_element
from ..version import EXCHANGE_2010_SP2
from .common import EWSService, folder_ids_element

log = logging.getLogger(__name__)


class ResolveNames(EWSService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/resolvenames-operation"""

    SERVICE_NAME = "ResolveNames"
    element_container_name = f"{{{MNS}}}ResolutionSet"
    ERRORS_TO_CATCH_IN_RESPONSE = ErrorNameResolutionNoResults
    WARNINGS_TO_IGNORE_IN_RESPONSE = ErrorNameResolutionMultipleResults
    # Note: paging information is returned as attrs on the 'ResolutionSet' element, but this service does not
    # support the 'IndexedPageItemView' element, so it's not really a paging service. According to docs, at most
    # 100 candidates are returned for a lookup.
    supports_paging = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.return_full_contact_data = False  # A hack to communicate parsing args to _elems_to_objs()

    def call(
        self,
        unresolved_entries,
        parent_folders=None,
        return_full_contact_data=False,
        search_scope=None,
        contact_data_shape=None,
    ):
        if self.chunk_size > 100:
            raise ValueError(
                f"Chunk size {self.chunk_size} is too high. {self.SERVICE_NAME} supports returning at most 100 "
                f"candidates for a lookup",
            )
        if search_scope and search_scope not in SEARCH_SCOPE_CHOICES:
            raise InvalidEnumValue("search_scope", search_scope, SEARCH_SCOPE_CHOICES)
        if contact_data_shape and contact_data_shape not in SHAPE_CHOICES:
            raise InvalidEnumValue("contact_data_shape", contact_data_shape, SHAPE_CHOICES)
        self.return_full_contact_data = return_full_contact_data
        return self._elems_to_objs(
            self._chunked_get_elements(
                self.get_payload,
                items=unresolved_entries,
                parent_folders=parent_folders,
                return_full_contact_data=return_full_contact_data,
                search_scope=search_scope,
                contact_data_shape=contact_data_shape,
            )
        )

    def _elem_to_obj(self, elem):
        if self.return_full_contact_data:
            mailbox_elem = elem.find(Mailbox.response_tag())
            contact_elem = elem.find(Contact.response_tag())
            return (
                None if mailbox_elem is None else Mailbox.from_xml(elem=mailbox_elem, account=None),
                None if contact_elem is None else Contact.from_xml(elem=contact_elem, account=None),
            )
        return Mailbox.from_xml(elem=elem.find(Mailbox.response_tag()), account=None)

    def get_payload(
        self, unresolved_entries, parent_folders, return_full_contact_data, search_scope, contact_data_shape
    ):
        attrs = dict(ReturnFullContactData=return_full_contact_data)
        if search_scope:
            attrs["SearchScope"] = search_scope
        if contact_data_shape:
            if self.protocol.version.build < EXCHANGE_2010_SP2:
                raise NotImplementedError(
                    "'contact_data_shape' is only supported for Exchange 2010 SP2 servers and later"
                )
            attrs["ContactDataShape"] = contact_data_shape
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=attrs)
        if parent_folders:
            payload.append(
                folder_ids_element(folders=parent_folders, version=self.protocol.version, tag="m:ParentFolderIds")
            )
        for entry in unresolved_entries:
            add_xml_child(payload, "m:UnresolvedEntry", entry)
        return payload



from ..protocol import BaseProtocol


class AutodiscoverProtocol(BaseProtocol):
    """Protocol which implements the bare essentials for autodiscover."""

    TIMEOUT = 10  # Seconds

    def __str__(self):
        return f"""\
Autodiscover endpoint: {self.service_endpoint}
Auth type: {self.auth_type}"""



from ..items import Persona
from ..properties import PersonaId
from ..util import MNS, create_element, set_xml_value
from .common import EWSAccountService, to_item_id


class GetPersona(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getpersona-operation"""

    SERVICE_NAME = "GetPersona"

    def call(self, personas):
        # GetPersona only accepts one persona ID per request. Crazy.
        for persona in personas:
            yield from self._elems_to_objs(self._get_elements(payload=self.get_payload(persona=persona)))

    def _elem_to_obj(self, elem):
        return Persona.from_xml(elem=elem, account=None)

    def get_payload(self, persona):
        return set_xml_value(
            create_element(f"m:{self.SERVICE_NAME}"), to_item_id(persona, PersonaId), version=self.protocol.version
        )

    @classmethod
    def _get_elements_in_container(cls, container):
        return container.findall(f"{{{MNS}}}{Persona.ELEMENT_NAME}")

    @classmethod
    def _response_tag(cls):
        return f"{{{MNS}}}{cls.SERVICE_NAME}ResponseMessage"



from ..properties import MovedItemId
from ..util import create_element
from .common import EWSAccountService, item_ids_element


class MarkAsJunk(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/markasjunk-operation"""

    SERVICE_NAME = "MarkAsJunk"

    def call(self, items, is_junk, move_item):
        return self._elems_to_objs(
            self._chunked_get_elements(self.get_payload, items=items, is_junk=is_junk, move_item=move_item)
        )

    def _elem_to_obj(self, elem):
        return MovedItemId.id_from_xml(elem)

    @classmethod
    def _get_elements_in_container(cls, container):
        return container.findall(MovedItemId.response_tag())

    def get_payload(self, items, is_junk, move_item):
        # Takes a list of items and returns either success or raises an error message
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=dict(IsJunk=is_junk, MoveItem=move_item))
        payload.append(item_ids_element(items=items, version=self.account.version))
        return payload



import logging

from cached_property import threaded_cached_property

from .credentials import BaseCredentials, OAuth2AuthorizationCodeCredentials, OAuth2Credentials
from .errors import InvalidEnumValue, InvalidTypeError
from .protocol import FailFast, RetryPolicy
from .transport import AUTH_TYPE_MAP, CREDENTIALS_REQUIRED, OAUTH2
from .util import split_url
from .version import Version

log = logging.getLogger(__name__)

DEFAULT_AUTH_TYPE = {
    # This type of credentials *must* use the OAuth auth type
    OAuth2Credentials: OAUTH2,
    OAuth2AuthorizationCodeCredentials: OAUTH2,
}


class Configuration:
    """Contains information needed to create an authenticated connection to an EWS endpoint.

    The 'credentials' argument contains the credentials needed to authenticate with the server. Multiple credentials
    implementations are available in 'exchangelib.credentials'.

    config = Configuration(credentials=Credentials('john@example.com', 'MY_SECRET'), ...)

    The 'server' and 'service_endpoint' arguments are mutually exclusive. The former must contain only a domain name,
    the latter a full URL:

        config = Configuration(server='example.com', ...)
        config = Configuration(service_endpoint='https://mail.example.com/EWS/Exchange.asmx', ...)

    If you know which authentication type the server uses, you add that as a hint in 'auth_type'. Likewise, you can
    add the server version as a hint. This allows to skip the auth type and version guessing routines:

        config = Configuration(auth_type=NTLM, ...)
        config = Configuration(version=Version(build=Build(15, 1, 2, 3)), ...)

    You can use 'retry_policy' to define a custom retry policy for handling server connection failures:

        config = Configuration(retry_policy=FaultTolerance(max_wait=3600), ...)

    'max_connections' defines the max number of connections allowed for this server. This may be restricted by
    policies on the Exchange server.
    """

    def __init__(
        self,
        credentials=None,
        server=None,
        service_endpoint=None,
        auth_type=None,
        version=None,
        retry_policy=None,
        max_connections=None,
    ):
        if not isinstance(credentials, (BaseCredentials, type(None))):
            raise InvalidTypeError("credentials", credentials, BaseCredentials)
        if auth_type is None:
            # Set a default auth type for the credentials where this makes sense
            auth_type = DEFAULT_AUTH_TYPE.get(type(credentials))
        if auth_type is not None and auth_type not in AUTH_TYPE_MAP:
            raise InvalidEnumValue("auth_type", auth_type, AUTH_TYPE_MAP)
        if credentials is None and auth_type in CREDENTIALS_REQUIRED:
            raise ValueError(f"Auth type {auth_type!r} was detected but no credentials were provided")
        if server and service_endpoint:
            raise AttributeError("Only one of 'server' or 'service_endpoint' must be provided")
        if not retry_policy:
            retry_policy = FailFast()
        if not isinstance(version, (Version, type(None))):
            raise InvalidTypeError("version", version, Version)
        if not isinstance(retry_policy, RetryPolicy):
            raise InvalidTypeError("retry_policy", retry_policy, RetryPolicy)
        if not isinstance(max_connections, (int, type(None))):
            raise InvalidTypeError("max_connections", max_connections, int)
        self._credentials = credentials
        if server:
            self.service_endpoint = f"https://{server}/EWS/Exchange.asmx"
        else:
            self.service_endpoint = service_endpoint
        self.auth_type = auth_type
        self.version = version
        self.retry_policy = retry_policy
        self.max_connections = max_connections

    @property
    def credentials(self):
        # Do not update credentials from this class. Instead, do it from Protocol
        return self._credentials

    @threaded_cached_property
    def server(self):
        if not self.service_endpoint:
            return None
        return split_url(self.service_endpoint)[1]

    def __repr__(self):
        args_str = ", ".join(
            f"{k}={getattr(self, k)!r}"
            for k in ("credentials", "service_endpoint", "auth_type", "version", "retry_policy")
        )
        return f"{self.__class__.__name__}({args_str})"



import logging
import time

import requests.auth
import requests_ntlm
import requests_oauthlib

from .errors import TransportError, UnauthorizedError
from .util import (
    CONNECTION_ERRORS,
    RETRY_WAIT,
    DummyResponse,
    _back_off_if_needed,
    _retry_after,
    add_xml_child,
    create_element,
    ns_translation,
    xml_to_str,
)

log = logging.getLogger(__name__)

# Authentication method enums
NOAUTH = "no authentication"
NTLM = "NTLM"
BASIC = "basic"
DIGEST = "digest"
GSSAPI = "gssapi"
SSPI = "sspi"
OAUTH2 = "OAuth 2.0"
CBA = "CBA"  # Certificate Based Authentication

# The auth types that must be accompanied by a credentials object
CREDENTIALS_REQUIRED = (NTLM, BASIC, DIGEST, OAUTH2)

AUTH_TYPE_MAP = {
    NTLM: requests_ntlm.HttpNtlmAuth,
    BASIC: requests.auth.HTTPBasicAuth,
    DIGEST: requests.auth.HTTPDigestAuth,
    OAUTH2: requests_oauthlib.OAuth2,
    CBA: None,
    NOAUTH: None,
}
try:
    import requests_gssapi

    AUTH_TYPE_MAP[GSSAPI] = requests_gssapi.HTTPSPNEGOAuth
except ImportError:
    # Kerberos auth is optional
    pass
try:
    import requests_negotiate_sspi

    AUTH_TYPE_MAP[SSPI] = requests_negotiate_sspi.HttpNegotiateAuth
except ImportError:
    # SSPI auth is optional
    pass

DEFAULT_ENCODING = "utf-8"
DEFAULT_HEADERS = {"Content-Type": f"text/xml; charset={DEFAULT_ENCODING}", "Accept-Encoding": "gzip, deflate"}


def wrap(content, api_version=None, account_to_impersonate=None, timezone=None):
    """Generate the necessary boilerplate XML for a raw SOAP request. The XML is specific to the server version.
    ExchangeImpersonation allows to act as the user we want to impersonate.

    RequestServerVersion element on MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/requestserverversion

    ExchangeImpersonation element on MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/exchangeimpersonation

    TimeZoneContent element on MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/timezonecontext

    :param content:
    :param api_version:
    :param account_to_impersonate:  (Default value = None)
    :param timezone:  (Default value = None)
    """
    envelope = create_element("s:Envelope", nsmap=ns_translation)
    header = create_element("s:Header")
    if api_version:
        request_server_version = create_element("t:RequestServerVersion", attrs=dict(Version=api_version))
        header.append(request_server_version)
    if account_to_impersonate:
        exchange_impersonation = create_element("t:ExchangeImpersonation")
        connecting_sid = create_element("t:ConnectingSID")
        # We have multiple options for uniquely identifying the user. Here's a prioritized list in accordance with
        # https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/connectingsid
        for attr, tag in (
            ("sid", "SID"),
            ("upn", "PrincipalName"),
            ("smtp_address", "SmtpAddress"),
            ("primary_smtp_address", "PrimarySmtpAddress"),
        ):
            val = getattr(account_to_impersonate, attr)
            if val:
                add_xml_child(connecting_sid, f"t:{tag}", val)
                break
        exchange_impersonation.append(connecting_sid)
        header.append(exchange_impersonation)
    if timezone:
        timezone_context = create_element("t:TimeZoneContext")
        timezone_definition = create_element("t:TimeZoneDefinition", attrs=dict(Id=timezone.ms_id))
        timezone_context.append(timezone_definition)
        header.append(timezone_context)
    if len(header):
        envelope.append(header)
    body = create_element("s:Body")
    body.append(content)
    envelope.append(body)
    return xml_to_str(envelope, encoding=DEFAULT_ENCODING, xml_declaration=True)


def get_auth_instance(auth_type, **kwargs):
    """Return an *Auth instance suitable for the requests package.

    :param auth_type:
    :param kwargs:
    """
    model = AUTH_TYPE_MAP[auth_type]
    if model is None:
        return None
    if auth_type == GSSAPI:
        # Kerberos auth relies on credentials supplied via a ticket available externally to this library
        return model()
    if auth_type == SSPI:
        # SSPI auth does not require credentials, but can have it
        return model(**kwargs)
    return model(**kwargs)


def get_service_authtype(service_endpoint, retry_policy, api_versions, name):
    # Get auth type by tasting headers from the server. Only do POST requests. HEAD is too error prone, and some servers
    # are set up to redirect to OWA on all requests except POST to /EWS/Exchange.asmx
    #
    # We don't know the API version yet, but we need it to create a valid request because some Exchange servers only
    # respond when given a valid request. Try all known versions. Gross.
    from .protocol import BaseProtocol

    retry = 0
    t_start = time.monotonic()
    headers = DEFAULT_HEADERS.copy()
    for api_version in api_versions:
        data = dummy_xml(api_version=api_version, name=name)
        log.debug("Requesting %s from %s", data, service_endpoint)
        while True:
            _back_off_if_needed(retry_policy.back_off_until)
            log.debug("Trying to get service auth type for %s", service_endpoint)
            with BaseProtocol.raw_session(service_endpoint) as s:
                try:
                    r = s.post(
                        url=service_endpoint,
                        headers=headers,
                        data=data,
                        allow_redirects=False,
                        timeout=BaseProtocol.TIMEOUT,
                    )
                    r.close()  # Release memory
                    break
                except CONNECTION_ERRORS as e:
                    # Don't retry on TLS errors. They will most likely be persistent.
                    total_wait = time.monotonic() - t_start
                    r = DummyResponse(url=service_endpoint, request_headers=headers)
                    if retry_policy.may_retry_on_error(response=r, wait=total_wait):
                        wait = _retry_after(r, RETRY_WAIT)
                        log.info(
                            "Connection error on URL %s (retry %s, error: %s). Cool down %s secs",
                            service_endpoint,
                            retry,
                            e,
                            wait,
                        )
                        retry_policy.back_off(wait)
                        retry += 1
                        continue
                    raise TransportError(str(e)) from e
        if r.status_code not in (200, 401):
            log.debug("Unexpected response: %s %s", r.status_code, r.reason)
            continue
        try:
            auth_type = get_auth_method_from_response(response=r)
            log.debug("Auth type is %s", auth_type)
            return auth_type, api_version
        except UnauthorizedError:
            continue
    raise TransportError("Failed to get auth type from service")


def get_auth_method_from_response(response):
    # First, get the auth method from headers. Then, test credentials. Don't handle redirects - burden is on caller.
    log.debug("Request headers: %s", response.request.headers)
    log.debug("Response headers: %s", response.headers)
    if response.status_code == 200:
        return NOAUTH
    # Get auth type from headers
    for key, val in response.headers.items():
        if key.lower() == "www-authenticate":
            # Requests will combine multiple HTTP headers into one in 'request.headers'
            vals = _tokenize(val.lower())
            for v in vals:
                if v.startswith("realm"):
                    realm = v.split("=")[1].strip('"')
                    log.debug("realm: %s", realm)
            # Prefer most secure auth method if more than one is offered. See discussion at
            # http://docs.oracle.com/javase/7/docs/technotes/guides/net/http-auth.html
            if "digest" in vals:
                return DIGEST
            if "ntlm" in vals:
                return NTLM
            if "basic" in vals:
                return BASIC
    raise UnauthorizedError("No compatible auth type was reported by server")


def _tokenize(val):
    # Splits cookie auth values
    auth_methods = []
    auth_method = ""
    quote = False
    for c in val:
        if c in (" ", ",") and not quote:
            if auth_method not in ("", ","):
                auth_methods.append(auth_method)
            auth_method = ""
            continue
        if c == '"':
            auth_method += c
            if quote:
                auth_methods.append(auth_method)
                auth_method = ""
            quote = not quote
            continue
        auth_method += c
    if auth_method:
        auth_methods.append(auth_method)
    return auth_methods


def dummy_xml(api_version, name):
    # Generate a minimal, valid EWS request
    from .services import ResolveNames  # Avoid circular import

    return wrap(
        content=ResolveNames(protocol=None).get_payload(
            unresolved_entries=[name],
            parent_folders=None,
            return_full_contact_data=False,
            search_scope=None,
            contact_data_shape=None,
        ),
        api_version=api_version,
    )



from ..errors import InvalidEnumValue, InvalidTypeError
from ..folders import BaseFolder
from ..items import (
    MESSAGE_DISPOSITION_CHOICES,
    SAVE_ONLY,
    SEND_AND_SAVE_COPY,
    SEND_MEETING_INVITATIONS_CHOICES,
    SEND_ONLY,
    BulkCreateResult,
)
from ..properties import FolderId
from ..util import MNS, create_element, set_xml_value
from .common import EWSAccountService, folder_ids_element


class CreateItem(EWSAccountService):
    """Take a folder and a list of items. Return the result of creation as a list of tuples (success[True|False],
    errormessage), in the same order as the input list.

    MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/createitem-operation
    """

    SERVICE_NAME = "CreateItem"
    element_container_name = f"{{{MNS}}}Items"

    def call(self, items, folder, message_disposition, send_meeting_invitations):
        if message_disposition not in MESSAGE_DISPOSITION_CHOICES:
            raise InvalidEnumValue("message_disposition", message_disposition, MESSAGE_DISPOSITION_CHOICES)
        if send_meeting_invitations not in SEND_MEETING_INVITATIONS_CHOICES:
            raise InvalidEnumValue(
                "send_meeting_invitations", send_meeting_invitations, SEND_MEETING_INVITATIONS_CHOICES
            )
        if folder is not None:
            if not isinstance(folder, (BaseFolder, FolderId)):
                raise InvalidTypeError("folder", folder, (BaseFolder, FolderId))
            if folder.account != self.account:
                raise ValueError("Folder must belong to account")
        if message_disposition == SAVE_ONLY and folder is None:
            raise AttributeError("Folder must be supplied when in save-only mode")
        if message_disposition == SEND_AND_SAVE_COPY and folder is None:
            folder = self.account.sent  # 'Sent' is default EWS behaviour
        if message_disposition == SEND_ONLY and folder is not None:
            raise AttributeError("Folder must be None in send-ony mode")
        return self._elems_to_objs(
            self._chunked_get_elements(
                self.get_payload,
                items=items,
                folder=folder,
                message_disposition=message_disposition,
                send_meeting_invitations=send_meeting_invitations,
            )
        )

    def _elem_to_obj(self, elem):
        if isinstance(elem, bool):
            return elem
        return BulkCreateResult.from_xml(elem=elem, account=self.account)

    @classmethod
    def _get_elements_in_container(cls, container):
        res = super()._get_elements_in_container(container)
        return res or [True]

    def get_payload(self, items, folder, message_disposition, send_meeting_invitations):
        """Take a list of Item objects (CalendarItem, Message etc) and return the XML for a CreateItem request.
        convert items to XML Elements.

        MessageDisposition is only applicable to email messages, where it is required.

        SendMeetingInvitations is required for calendar items. It is also applicable to tasks, meeting request
        responses (see
        https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/createitem-operation-meeting-request
        ) and sharing
        invitation accepts (see
        https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/createitem-acceptsharinginvitation
        ). The last two are not supported yet.

        :param items:
        :param folder:
        :param message_disposition:
        :param send_meeting_invitations:
        """
        payload = create_element(
            f"m:{self.SERVICE_NAME}",
            attrs=dict(MessageDisposition=message_disposition, SendMeetingInvitations=send_meeting_invitations),
        )
        if folder:
            payload.append(
                folder_ids_element(folders=[folder], version=self.account.version, tag="m:SavedItemFolderId")
            )
        item_elems = create_element("m:Items")
        for item in items:
            if not item.account:
                item.account = self.account
            set_xml_value(item_elems, item, version=self.account.version)
        payload.append(item_elems)
        return payload
